/************************************************************************/
/*                                                                      
  	BANCOMM bc63{57}VME/bc350VXI Time and Frequency Processor Driver.
  	
  	for HP-UX 9.01 Systems
                                                                      
	Copyright (C) 1994, BANCOMM, San Jose, CA.  All rights reserved.	
                                                                      
  	Author:	Greg Dowd	
*/                                                                      
/****************************************************************************/
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/ioctl.h>
#include "/usr/conf/wsio/vme2.h"
#include "/usr/conf/h/conf.h"
#include "/usr/conf/h/wsio.h"
#include "/etc/conf/h/io.h"
#include <stdio.h>

#include "btfpd.hpux.h"				/* All needed headers and definitions. 	*/

/****************************************************************************/
/*	btfp device state information structures.  MAXDEV is defined in
	btfpd.h as the max number of bc637's this driver can support.
*/
#define BTFP_STRING     "BTFP VME 10.20 driver"

static struct btfp_status {	
		int intwait;				/* Copy of MASK interrupt enable bits.*/
		int	signalon;				/* True if signal sending is enabled.*/
		int	packeton;				/* True if packet mode is enabled.*/
		int	debugon;				/* True if debug mode is enabled.*/
		struct proc *procp;			/* Pointer to calling process */
		int gps_synch;				/* True if synched with GPS packets */
		int rpacki;					/* Read Packet buffer index.			*/
		int wpacki;					/* Write Packet buffer index.			*/
		unsigned char rpack_buffer[MAXPACKET+1];/* used for read calls */
		unsigned char wpack_buffer[MAXPACKET+1];/* used for write calls */
		unsigned char work_buffer[MAXPACKET+1];	/* used to build packet */
		unsigned char vector;					/* btfp vme interrupt vector */
		unsigned char level;					/* btfp vme interrupt level */
} btfp_status[ MAXDEV ];			/* One for each bc63{57} supported.	 */

struct isc_table_type *btfp_isc[MAXDEV]; /* Misc entries for each board */

BTFPVOLATILE *btfp_vme_addr[MAXDEV];/* Device address info from header 	*/
BTFPVOLATILE *btfp_addr[MAXDEV];    /* Virtual address info				*/
int	 btfp_vme_int[MAXDEV];			/* Device interrupt info from header	*/
char frsttime[MAXDEV];				/* One time var to init vec & level	*/

int btfp_intr();					/* Proto for interrupt routine		*/
int btfp_intr1();

int btfp_attach();					/* Proto for attach routine			*/
int (*btfp_saved_attach)();			/* linked list for vme_attach 		*/
/************************************************************************/
/*
	btfp_attach - called during power-up

					Try to access each btfp in the SYSCONFIG file, up to
					MAXDEV, which is our local definition
					of how many btfp's this driver can support.
					If an access succeeds, mark the board		
					on-line and announce it is there, else mark it		
					off-line and announce it is not there.  'Opens' to 
					boards either not there or beyond MAXDEV will fail.
					
					For boards that are there, their interrupt registers
					are loaded so that a subsequent interrupt enable (via
					a WMASK ioctl(2) call which turns on any bits) will
					allow an interrupt to come in.
*/
/************************************************************************/
btfp_attach(id, isc)
char *id;
struct isc_table_type *isc;
{
	register BTFPVOLATILE *dp;				/* Pointer to board base reg.	*/
	register num_dev;						/* # devices to initialize.		*/
	register i;								/* Index to a btfp & strucs.	*/
	unsigned char my_vector;				/* rtn_value from vme_isrlink	*/

	num_dev = 0;
#ifdef BTFP0_VME_ADDR
	btfp_vme_addr[0] = BTFP0_VME_ADDR;
	btfp_vme_int[0]  = BTFP0_VME_INT;
	num_dev++;
#endif
#ifdef BTFP1_VME_ADDR
	btfp_vme_addr[1] = BTFP1_VME_ADDR;
	btfp_vme_int[1]  = BTFP1_VME_INT;
	num_dev++;
#endif
#ifdef BTFP2_VME_ADDR
	btfp_vme_addr[2] = BTFP2_VME_ADDR;
	btfp_vme_int[2]  = BTFP2_VME_INT;
	num_dev++;
#endif
#ifdef BTFP3_VME_ADDR
	btfp_vme_addr[3] = BTFP3_VME_ADDR;
	btfp_vme_int[3]  = BTFP3_VME_INT;
	num_dev++;
#endif

        if (strcmp(id, BTFP_STRING)) /* is it ours? */
                return; /* nope */
        isc_claim(isc,NULL);
	/* Initialize all boards and their structures.							*/
	for (i = 0; i < num_dev; i++) 			
	{	
		if((btfp_isc[i] = vme_create_isc()) == NULL)
		{
			printf("btfpinit(): Failed to allocate isc %d\n",i);
			return(ENXIO);
		}
		VME_INIT_IF_INFO(btfp_isc[i],D08_IRQ_TYPE,VME_A16);
											/* initialize isc->if_info		*/
		if((btfp_addr[i] = (BTFPVOLATILE *) map_mem_to_host
		  				   (btfp_isc[i],btfp_vme_addr[i],64)) == NULL)
		{
			free_isc(btfp_isc[i]);
			btfp_addr[i] = 0;
			printf("btfp_init: Failed to map btfp%d\n",i);
		}
		else
		{
			if (vme_testr(btfp_isc[i],(caddr_t)btfp_addr[i],SHORT_WIDE,0) <=0)
			{
				unmap_mem_from_host(btfp_isc[i],btfp_addr[i],64);
				free_isc(btfp_isc[i]);
				printf("btfpinit(): btfp%d is not accessible at %x.\n",
														i, btfp_addr[i]);
				btfp_addr[i] = 0;
			}
			else
			{
				printf ("btfpinit(): btfp%d  is accessible at %x.\n",
														i, btfp_addr[i]);
											/* Announce module was found.	*/
				
				my_vector = vme_isrlink(btfp_isc[i],btfp_intr,0x0,NULL,0x0);
				if(!(my_vector))
				{
					printf("btfpinit(): error setting interrupt handler!\n");
					return(ENXIO);
				}
				
				frsttime[ i ] = 1;				/* Set 1rst open to init.	*/
				btfp_isc[i]->if_drv_data = my_vector;

				btfp_status[ i ].intwait  = 0;/* Indicate not waiting for int.*/
				btfp_status[ i ].signalon = 0;/* Signal mechanism turned off.*/
				btfp_status[ i ].packeton = 0;/* Packet mechanism turned off.*/
				btfp_status[ i ].debugon  = 0;/* Debug mode is turned off.	*/
				btfp_status[ i ].gps_synch= 0;/* Not synched with GPS packets.*/
				btfp_status[ i ].vector	  = my_vector;
				btfp_status[ i ].level    = btfp_vme_int[i];
			
			} /* end of vme_testr routine */

		} /* end of map_mem_to_host */
	
	} /* end of i loop */

} /* end of btfpinit routine */

/************************************************************************/
/*                                                                     
  	btfp_open()- Called when an appication calls open(2).                                            
  								                                    
  					If the incoming device minor number which is an index
  					to all of the btfpn's is larger than MAXDEV which
  					was set by config, or if the btfp_init routine did
  					not find this bc63{57}, then the open fails, else 
					it succeeds.
  					
  					The module's MASK register is written to 
  					disable external interrupts, the control register 0
					is set to disable interrupts, and the driver's
  					indicators for this device are initialized
  					accordingly.
 
*/                                                                      
/************************************************************************/

btfp_open( mmdev, flags )
	dev_t mmdev;						/* Device (major/minor) number.	*/
	int flags;							/* Read/write flag. 			*/
{
	register BTFPVOLATILE *dp = btfp_addr[ minor(mmdev) ];
   										/* Init pointer to base register.*/
	
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/
	register i;
	ushort		ustemp, usmask;			/* For use turning off ints.	*/
	int		s;							/* Interrupt level from spl().	*/

	if(frsttime[minor(mmdev)])
	{
		dp->btfp_level  = sp->level;
		dp->btfp_vector = sp->vector;
		frsttime[minor(mmdev)] = 0;
	}
	
	if( ( minor(mmdev) >= MAXDEV ) || (dp == (BTFPVOLATILE *)0 ) ) 
	{
		printf("minor(mmdev) = %d\n",minor(mmdev));
		printf("dp = %4x\n",dp);
		printf("btfp_open: no such device or address error\n");
		u.u_error = ENXIO;
		return;				/* Set no such device or address err.*/
	}

	/* Do a successful, normal, open...	*/
	s = spln( btfp_status[minor (mmdev)].level );	/* Block interrupts */
	sp->intwait = 0;					/* Indicate not waiting for an int.*/
	sp->signalon  = 0;					/* Signal mechanism is turned off.	*/
	sp->packeton  = 0;					/* Packet mode is turned off.	*/
	sp->debugon  = 0;					/* Debug mode is turned off.	*/

	dp->btfp_mask = 0;					/* Turn off all interrupts.		*/

	usmask = LOCKDIS | HBDIS | EVENTDIS | STRDIS; 
										/* Disable lockout, heartbeat */
										/* Event, and strobe interrupts */
	dp->btfp_cr0 = usmask;
 	
	sp->procp = u.u_procp;			/* From the u structure (in case of kill)*/
	splx( s );							/* Turn interrupts back on.			*/
	return( 0 );						/* Else return open successful.		*/
} /* end of btfpopen routine */

/************************************************************************/
int
spln( ipl )
	int ipl;							/* Processor prio level to set.		*/
{
	switch ( ipl )						/* Call appropriate spln() based on	*/
	{									/*	the incoming argument.			*/
	case 2:								/* This is the most likely level.	*/
		return( spl2() );
		break;
	case 0:
		return( spl0() );
		break;
	case 1:
		return( spl1() );
		break;
	case 3:
		return( spl3() );
		break;
	case 4:
		return( spl4() );
		break;
	case 5:
		return( spl5() );
		break;
	case 6:
		return( spl6() );
	default:
		return( spl2() );			/* Probable best ipl for this module.	*/
		break;
	}
} /* end of spln routine */


/************************************************************************/
/*                                                                      
  	btfp_close()- Called on the last close(2) system call to the device.                                            
  							                                       
  					This routine turns off interrupts from all 5 interrupt 
  					sources on the bc63{57}, so none will occur when the
  					device is not in use.  It resets the driver's status
  					indicators accordingly, too.  Remember, close is only
  					called on the LAST device to close of all that are
  					open.

*/                                                                      
/************************************************************************/


btfp_close( mmdev, flag )
	dev_t mmdev;                        /* Major & minor device numbers.*/
	int flag;                          	/* Read/write flag. 			*/
{
	register BTFPVOLATILE *dp = (BTFPVOLATILE *) btfp_addr[ minor(mmdev) ];
   										/* Init pointer to base register.*/
	
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/
	unsigned short	ustemp, usmask;		/* For use turning off ints.	*/
	int		s;							/* Interrupt level from spl().	*/

	
	s = spln( btfp_status[minor(mmdev)].level );	/* Block interrupts */
	sp->intwait = 0;					/* Indicate not waiting for an int.*/
	sp->signalon  = 0;					/* Signal mechanism is turned off.	*/
	sp->packeton  = 0;					/* Packet mode is turned off.	*/
	sp->gps_synch  = 0;					/* Not synched with GPS packets. */

	dp->btfp_mask = 0;					/* Turn off all 5 interrupts.		*/

	usmask = LOCKDIS | HBDIS | EVENTDIS | STRDIS; 
										/* Disable lockout, heartbeat */
										/* Event, and strobe interrupts */
	dp->btfp_cr0 = usmask;

	sp->procp = 0;						/* No processes to signal on last close.*/
	splx( s );							/* Turn interrupts back on.			*/
	return( 0 );						/* Return close successful.			*/
} /* end of btfpclose routine */

/************************************************************************/
/*
	send_packet - send a packet to the INFIFO.

		NOTE: packet must be null-terminated!!!!

		References MAX_SPIN to avoid infinite loop.

		NOTE: interrupts should be disabled in this routine. Since we
			  call it from btfpintr and elsewhere, we depend on the
			  caller to block the interrupts.
*/
/************************************************************************/
int
send_packet (mmdev, packet)
dev_t mmdev;
unsigned char *packet;
{
	register BTFPVOLATILE *dp = (BTFPVOLATILE *) btfp_addr[ minor(mmdev) ];
   										/* Init pointer to base register.*/
	/* Pointer to btfp registers. 	*/
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/
	int loop_count;
	unsigned char *byte_ptr=packet;
	unsigned short ustemp;				/* For accessing ACK register */
	int		s;							/* Interrupt level from spl().	*/

	dp->btfp_ack = ACK_INFIFO;			/* Clear acknowledge bit */

	while (*byte_ptr)
	{
		dp->btfp_infifo = *byte_ptr;
		byte_ptr++;
	} /* end of while */


	dp->btfp_ack = ACK_INACT;			/* "Take-Action" bit */

										/* Wait for bc63{57} to take action */

										/* Note: no interrupt generated, */
										/* so we have to spin */

	loop_count = 0;
	while (loop_count <= MAX_SPIN)
	{
		ustemp = dp->btfp_ack;
		if ( ustemp & ACK_INFIFO ) /* Break if action taken */
		{
			break;
		}
		else
		{
			loop_count++;
		}
	}

	dp->btfp_ack = ACK_INFIFO;			/* Clear acknowledge bit */

	if (loop_count >= MAX_SPIN)
	{
		printf ("TIMEOUT (%d) waiting for INFIFO action\n", MAX_SPIN);
		return (1);								/* Return Error */
	}
	else
	{
		if (sp->debugon)
			printf ("%dLoop ", loop_count);
	}

	return (0);									/* Return No Problem */

} /* end of send_packet routine */

/************************************************************************/
/*                                                                      
  	btfp_read()- Called when an application issues a read(2) call.                                             
  						                                            
		(default) PACKETOFF mode:
                                                                      
		Set to Free Running Mode (mode 1), then makes a time request.

  		Access the bc63{57}'s capture register to freeze the time,		    
  		then return (converting BCD to ASCII) bytes of the time			
		to the caller, until the number of bytes he requested is		
		reached or until we run	out of bytes to return.  				

		Sends 34 bytes maximum, separating each of the 16 bytes			
		of time data with 3 spaces, and new-line and null 				
		terminating the string.											
												    					
      	This routine exits when 34 bytes have been sent in response
		to a read, regardless of what uio.uio_iov.iov_len is.  If it was
		not zero, then presumably the caller (e.g. cat()) will call
		again, but with uio.uio_resid = 33, which is our indication to
		just return without any transfer, which will result in the
		caller seeing 0 bytes transferred, which means EOF, by
		UNIX definition.

		SIGNALON/SIGNALOFF modes have no effect on this mode.

		PACKETON mode:

		Block until a data packet is available from the OUTFIFO (unless
		SIGNALON mode), then fill the buffer supplied by the read call
		with the data from the OUTFIFO.  
		
			SIGNALON notes:

			If both PACKETON and SIGNALON modes are in effect, then
			the user must issue the ioctl calls to clear the
			DPA interrupt status (WINTSTAT) and enable DPA interrupts
			in the MASK register in a SIGNALON (non-blocking) WMASK ioctl call.
			When the signal is received to indicate the DPA interrupt has
			occurred, the signal handler should issue the read(2) call,
			still leaving the driver in the SIGNALON mode. This will
			cause the btfpread() routine to immediately read the data
			packet from the OUTFIFO.

*/
/************************************************************************/

#define	MAXTBYTES	49						/* Size of returned time string.*/
#define	MAXABYTES	22						/* Size of converted time array.*/

btfp_read( mmdev,uio )
	dev_t mmdev;							/* Major & minor device info. 	*/
	struct uio *uio;
{
	register BTFPVOLATILE *dp = btfp_addr[ minor(mmdev) ];
   										/* Init pointer to base register.*/
	
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/
   	unsigned short dummy;					/* Dummy for use accessing captr.*/
   	unsigned short ustemp, ustemp2;			/* Temps for use accessing regs.*/
   	char timearray[ MAXABYTES ];			/* Place to put converted time.	*/
   	register i = 0;							/* Loop index thanks to K&R.	*/
    register twi;							/* Time word index.				*/
   	register tai;							/* Time array index.			*/
   	unsigned char lochar;					/* Place to build the ASCII.	*/
   	unsigned char hichar;					/* Ditto.						*/
	int rtn;								/* catcher for return code */
	int s;									/* return from spl()	*/
	int length;								/* Used for PACKETON read		*/
	unsigned char *user_buffer;				/* pointer into user buffer */
   	char space[3];              			/* blank spaces for readability */	
   	union char_union 					/* For converting short to 'chars'.*/
   	{
   			unsigned char	two_chars[ 2 ];	/* Two characters holder.		*/
   			unsigned short	time_word;		/* One of the time words.		*/
   	} ch_union;

  	for(i=0;i<3;i++)space[i]=' ';			/* Initialize spaces	*/ 	
	i=0;
	if( minor(mmdev) >= MAXDEV )			/* Beyond supportable range.	*/
	{
		u.u_error = ENXIO;			 		/* Return error indication.	*/   	
		return;
	}	
	if (!sp->packeton)
	{
	/* PACKETOFF mode (default) */

	dummy = dp->btfp_unlock;				/* Release capture lock if set.	*/
	dummy = dp->btfp_timereq;				/* Freeze the time via access to*/
   											/*	the timereq register.		*/
   											
	/* Unpack bytes from btfp..	*/   												
   	timearray[ MAXABYTES - 1 ] = '\0';		/* Set up zero termination.		*/
   	timearray[ MAXABYTES - 2 ] = '\n';		/* And new-line termination.	*/											

	tai = MAXABYTES - 3;					/* Fill time array back to front.*/
	for( twi = 4; twi >= 0; twi--)			/* Read time; BCD->ASCII.		*/
	{
	   	ch_union.time_word = dp->btfp_time[ twi ];
											/* Get a time word from module*/
		lochar = ch_union.two_chars[ 1 ];	/* Get two digits to process.	*/
		hichar = ch_union.two_chars[ 1 ];	/* And into another var.		*/
		
		lochar = ( lochar >> 4 ) | 0x30;	/* Convert a nibble to ASCII byte.*/
		hichar = ( hichar & 0xf) | 0x30;	/* Convert the other nibble.	*/
		
		timearray[ tai ] = hichar;			/* Move into the time array.	*/
		tai--;								/* Decrement the time array indx.*/
		timearray[ tai ] = lochar;			/* Move the other byte.			*/
		tai--;								/* Decrement the time array indx.*/

		lochar = ch_union.two_chars[ 0 ];	/* Get two digits to process.	*/
		hichar = ch_union.two_chars[ 0 ];	/* And into another var.		*/
		
		lochar = ( lochar >> 4 ) | 0x30;	/* Convert a nibble to ASCII byte.*/
		hichar = ( hichar & 0xf) | 0x30;	/* Convert the other nibble.	*/
		
		timearray[ tai ] = hichar;			/* Move into the time array.	*/
		tai--;								/* Decrement the time array indx.*/
		timearray[ tai ] = lochar;			/* Move the other byte.			*/
		tai--;								/* Decrement the time array indx.*/
		if( tai == 0 ) break;				/* End the loop after MAXTBYTES. */   	
   	}
	i = 0;									/* Initialize byte index.		*/

   	while ( uio->uio_iov->iov_len )    		/* While application wants chars.*/
	{
		if( (i <= MAXABYTES - 1) && 		/* Deliver up to max bytes...	*/
			(uio->uio_offset < MAXTBYTES) )	/* And stop on a later read.	*/
		{
			if( i == 2 ||					/* Between the time fields...	*/
				i == 3 ||					/* Day, hour, minute..etc...	*/
				i == 6 ||				
				i == 8 ||
				i == 10 ||
				i == 12 ||
				i == 15 ||
				i == 18 ||
				i == 19)					/* Insert spaces for readability.*/
			{
				uiomove(space,3,UIO_READ,uio); /* transfer spaces to user */	
			}
			uiomove(&timearray[ i ],1,UIO_READ,uio);
											/* Move time byte to user space.*/
			i++;							/* Move on to the next byte.	*/
		}
		else								/* We have no more to give...	*/
		{									/* Or he doesn't want any more.	*/
			break;							/* Exit the while and the read.	*/
		}
   	}
	return( 0 );							/* Return no problem.			*/

	} /* end of PACKETOFF mode */

	else /* start of PACKETON mode */
	{
		if( uio->uio_iov->iov_len >= sizeof (sp->rpack_buffer) )
											/* Beyond packet buffer size.	*/
		{
			u.u_error = ENXIO;			 	/* Return error indication.		*/   	
			return;
		}
		if (! (sp->signalon) )				/* Only block in SIGNALOFF mode */
		{

			s = spln( btfp_status[minor (mmdev)].level );
											/* Block interrupts */
											/* Wait for Data Packet Available */
				if ( !(sp->gps_synch) )		/* Are we in synch with GPS? */
				{
					dp->btfp_ack = ACK_CLEAR;		/*  Clear OUTFIFO		*/
													/* When we get the next */
													/* DPA, the interrupt   */
													/* routine will get us  */
													/* synch-ed up with GPS */
													/* and set the flag     */
				}
				dp->btfp_ack = ACK_DPA;
				sp->intwait = MASK_DPA;				/* Indicate waiting for a DPA int.*/
				dp->btfp_mask = MASK_DPA;			/* Set MASK reg. for a DPA int.*/

				while ( sp->intwait )   /* btfpintr() will clear this on int.	*/
				{						/* Block, waiting for a bc63{57} interrupt.*/
					sleep( (caddr_t)&sp->intwait, PZERO + 5 ); 
										/* Sleep till our interrupt occurs.		*/
				}

				splx( s );				/* Allow our device interrupts again.	*/
		} /* end of SIGNALOFF section */

		/* Interrupt routine has synched us up with the TANS and read in */
		/* the next packet (rpacki bytes) from the OUTFIFO into rpack_buffer */

		if (sp->rpacki <= uio->uio_iov->iov_len)
										/* If less bytes than requested */
		{
			uiomove(sp->rpack_buffer,sp->rpacki,UIO_READ,uio);	
		} /* end of "emptied our buffer first" case */
		else
		{
		  /* user didn't give us a big enough buffer! */
			uiomove(sp->rpack_buffer,uio->uio_iov->iov_len,UIO_READ,uio);
		} /* end of "filled user buffer" case */

		return (0);							/* return No Problem */
	} /* end of PACKETON mode */

} /* end of btfpread routine */

/************************************************************************/
/*
	btfp_write() - Called when an application does a write(2) call.

	(default) PACKETOFF mode:

	Sets to Free Running Mode (mode 1), waits for the 1PPS pulse (unless
	in SIGNALON mode), then loads time from user buffer.

	Accepts a buffer in UNIX's "date '%j%H%M%S'" format, namely Julian date,
	hours, minutes, then seconds (with or without intervening spaces). Prepends
	this buffer with a 'B' (to indicate Load Major Time). Note that this routine
	can accept the same format output by the default read(2) call, by
	stripping off the ms, us, and ns fields.
		
			SIGNALON notes:

			If both PACKETOFF and SIGNALON modes are in effect, then
			the user must issue the ioctl calls to clear the
			1PPS interrupt status (WINTSTAT) and enable 1PPS interrupts
			in the MASK register in a SIGNALON (non-blocking) WMASK ioctl call.
			When the signal is received to indicate the 1PPS interrupt has
			occurred, the signal handler should issue the write(2) call,
			still leaving the driver in the SIGNALON mode. This will
			cause the btfpwrite() routine to immediately write the data
			packet (containing the date and time) to the INFIFO.

	PACKETON mode:

	Accepts a buffer from the user and transmits it directly to the INFIFO.

	SIGNALON/SIGNALOFF modes have no effect on this mode.

/************************************************************************/

btfp_write(mmdev,uio)
	dev_t mmdev;							/* Major & minor device info.*/
	struct uio *uio;
{
	register BTFPVOLATILE *dp = btfp_addr[ minor(mmdev) ];
   										/* Init pointer to base register.*/
	
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/

	int rtn;								/* catcher for return code */
	int s;									/* spl() rtn code */
	unsigned char *user_buffer;				/* pointer into user buffer */
	register worki;							/* buffer index */
	int	num_bytes;							/* Number of user bytes */
	unsigned short usmask;					/* Used for register access */

	if( minor(mmdev) >= MAXDEV )			/* Beyond supportable range.	*/
	{
		u.u_error = ENXIO;			 		/* Return error indication.		*/   	
		return;
	}
	if (!sp->packeton)
	{
	/* PACKETOFF mode (default) */

		num_bytes = uio->uio_iov->iov_len;	/* Save number of user bytes */
											/* Do this now to save time later */
		uiomove(sp->work_buffer,uio->uio_iov->iov_len,UIO_WRITE,uio);
											/* Leave room for 'B' */
		sp->wpack_buffer[0] = SOH;			/* Make into null-terminated B packet */
		sp->wpack_buffer[1] = 'B';			
		sp->wpacki = 2;
		for (worki=0; worki <= num_bytes; worki++ )
											/* Eliminate white space, plus */
											/* ms, us, and ns fields */
		{
			if ( (sp->work_buffer[worki] != ' ')
				&& (sp->work_buffer[worki] != '\n')
				&& (sp->wpacki <= 10) ) 		/* (dddhhmmss takes 9 bytes) */
			{
				sp->wpack_buffer[sp->wpacki++] = sp->work_buffer[worki];
			}
		}
		sp->wpack_buffer[sp->wpacki++] = ETB;			/* Finish it off */
		sp->wpack_buffer[sp->wpacki] = (char) NULL;	/* with ETB and NULL */

											/* Wait for 1PPS to come around */

		if (! sp->signalon)					/* only block in SIGNALOFF mode */
		{


		s = spln( btfp_status[minor (mmdev)].level );	/* Block interrupts */
		if (sp->debugon)
			printf ("T");
		sp->intwait = MASK_1PPS;				/* Indicate waiting for a 1PPS int.*/
		dp->btfp_intstat = INTSTAT_1PPS;		/* Clear 1PPS int. status */
		usmask = dp->btfp_intstat;				/* Clear ALL int. status */
		dp->btfp_intstat = usmask;				/* ... */
		dp->btfp_mask = MASK_1PPS;				/* Set MASK reg. for a 1PPS int.*/
		while ( sp->intwait )   /* btfpintr() will clear this on int.	*/
		{						/* Block, waiting for a bc63{57} interrupt.*/
			sleep( (caddr_t)&sp->intwait, PZERO + 5 ); 
								/* Sleep till our interrupt occurs.		*/
		}
		splx( s );				/* Allow our device interrupts again.	*/

		/* Interrupt routine should send the packet in wpackbuffer */ 

		} /* end of (blocking) SIGNALOFF mode */


		s = spln( btfp_status[minor (mmdev)].level );	/* Block interrupts */
		if ( (send_packet (mmdev, sp->wpack_buffer)) == 1)	/* Load time buffer */
		{
			splx( s );				/* Allow our device interrupts again.	*/
			printf ("btfpwrite: Error from send_packet of wpack_buffer\n");
			return (EIO);
		}
		splx( s );				/* Allow our device interrupts again.	*/


		return( 0 );						/* Return no problem.			*/
	}
	else
	{
	/* PACKETON mode */

		if( uio->uio_iov->iov_len >= sizeof (sp->wpack_buffer) )
											/* Beyond packet buffer size.	*/
		{
			printf("btfp_write:Packet TOO large\n");
			return;
		}
		num_bytes = uio->uio_iov->iov_len;	/* Save number of user bytes */

		uiomove(sp->wpack_buffer,uio->uio_iov->iov_len,UIO_WRITE,uio);

		sp->wpack_buffer[num_bytes] = (unsigned char) NULL;
									/* Make into null-terminated packet */
		s = spln( btfp_status[minor (mmdev)].level );	/* Block interrupts */

		if ( (send_packet (mmdev, sp->wpack_buffer)) == 1)
													/* Send raw packet */
		{
			splx( s );					/* Turn interrupts back on.			*/
			return (EIO);
		}
		splx( s );						/* Turn interrupts back on.			*/

		return( 0 );							/* Return no problem.		*/
	} /* end of PACKETON mode */

} /* end of btfpwrite routine */

/************************************************************************/
/*                                                                      
  	btfp_ioctl()- Called when an application does an ioctl(2) call.                                           
  								                                   
  					Generally, the bc63{57}'s registers are written or read
  					by these routines.
					
					If we are in SIGNALON mode, we do not block when
					an external interrupt enable (to MASK) is received,
					else we do block.

                    /// We are enabling the interrupt after the cause
                    	of it may have happened, e.g. strobe may already
                    	have gone off.  Hence an application would sleep
                    	forever.                                               
*/                                                                   
/************************************************************************/

btfp_ioctl( mmdev, cmd, arg, mode )
	dev_t mmdev;                    /* Major & minor device info.			*/
	int cmd;                        /* The command. 						*/
	int *arg;                   	/* The data argument or pointer to it.	*/
	int	mode;						/* Unused in this driver.				*/
{
	register BTFPVOLATILE *dp = (BTFPVOLATILE *) btfp_addr[ minor(mmdev) ];
   										/* Init pointer to base register.*/
	
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/
	register struct btfp_device *up;
									/* READALL receives such a pointer.		*/
	register struct btfp_time *tp;
									/* Pointer to time register structure.	*/
	register struct btfp_event *ep;
									/* Pointer to event register structure.	*/
	register struct btfp_strobe *strp;
									/* Pointer to strobe register structure.*/
	struct btfp_device btfpdev;
	struct btfp_time btfptime;
	struct btfp_event btfpevent;
	struct btfp_strobe btfpstrobe;

	int		s;						/* Interrupt level from spl().			*/
   	unsigned short ustemp, ustemp2;	/* Temps for use accessing regs			*/
   	unsigned int uitemp;			/* Temp for use in returning values.	*/
	int i;							/* Temp loop variable.					*/	
	unsigned short 	usmask = 
					MASK_INT0 | MASK_INT1 | MASK_INT2 | MASK_INT3 | MASK_INT4;
									/* Mask for turning off device interrupts.*/ 

	up = &btfpdev;			
	tp = &btfptime;			
	ep = &btfpevent;			
	strp = &btfpstrobe;			
	
	switch (cmd) {					/* Handle specific incoming commands.	*/
	
	case RIDR:						/* Read the ID register.		*/
   		ustemp = dp->btfp_id;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;	
		break;
	
	case RDTR:						/* Read the Device Type register.		*/
   		ustemp = dp->btfp_device;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;
	
	case RSR:						/* Read the Status register.			*/
   		ustemp = dp->btfp_status;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case WCR:						/* Write to the Control register.		*/
   		dp->btfp_control = (ushort)*arg;
									/* Directly from the user's argument.	*/
		break;
	
	case RTIMEREQ:					/* Read the Time request register.		*/
   		ustemp = dp->btfp_timereq;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case RTIME0:					/* Read time word zero.					*/
		ustemp = dp->btfp_time[ 0 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case RTIME1:					/* Read time word one.					*/
		ustemp = dp->btfp_time[ 1 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case RTIME2:					/* Read time word two.					*/
		ustemp = dp->btfp_time[ 2 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case RTIME3:					/* Read time word three.				*/
		ustemp = dp->btfp_time[ 3 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case RTIME4:					/* Read time word four.					*/
		ustemp = dp->btfp_time[ 4 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case REVENT0:					/* Read event word zero.				*/
		ustemp = dp->btfp_event[ 0 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case REVENT1:					/* Read event word one.					*/
		ustemp = dp->btfp_event[ 1 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case REVENT2:					/* Read event word two.					*/
		ustemp = dp->btfp_event[ 2 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case REVENT3:					/* Read event word three.				*/
		ustemp = dp->btfp_event[ 3 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case REVENT4:					/* Read event word four.				*/
		ustemp = dp->btfp_event[ 4 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case WSTROBE1:					/* Write strobe word one.				*/
   		dp->btfp_strobe1 = (ushort)*arg;	
									/* Directly from the user's argument.	*/
		break;

	case WSTROBE2:					/* Write strobe word two.				*/
   		dp->btfp_strobe2 = (ushort)*arg;
									/* Directly from the user's argument.	*/
		break;

	case WSTROBE3:					/* Write strobe word three.				*/
   		dp->btfp_strobe3 = (ushort)*arg;
									/* Directly from the user's argument.	*/
		break;

	case RUNLOCK:					/* Read Unlock register.				*/
   		ustemp = dp->btfp_unlock;	/* Contents of the register matter not.	*/
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		break;
	
	case RACK:						/* Read the ACK register.				*/
   		ustemp = dp->btfp_ack;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		*arg   = uitemp;
		break;

	case WACK:						/* Write to the ACK register.			*/
   		dp->btfp_ack = (ushort)*arg;/* Directly from the user's argument.	*/
		break;

	case WCR0:						/* Write control register zero.		 	*/
		dp->btfp_cr0 = (ushort)*arg;/* Directly from the user's argument.	*/
		break;

	case RCR0:						/* Read control register zero.		 	*/
		ustemp = dp->btfp_cr0;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		*arg   = uitemp;
		break;
	
	case ROUTFIFO:					/* Read the OUTFIFO register.			*/
   		ustemp = dp->btfp_outfifo;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		*arg   = uitemp;
		break;

	case WINFIFO:					/* Write to the INFIFO register.		*/
   		dp->btfp_infifo = (ushort)*arg;/* Direct from the user's argument.	*/
		break;
	
	case RMASK:						/* Read the MASK register.				*/
   		ustemp = dp->btfp_mask;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		*arg   = uitemp;
		break;

	case WMASK:						/* Write to the MASK register.			*/
		if( !( (int)*arg & ( (int) usmask ) ) )
		{
			dp->btfp_mask = (ushort)*arg; 
									/* Direct from the user's argument.		*/
									/* No interrupt enable requested.		*/
		}
		else if( !sp->signalon )	/* Signal sending is NOT enabled.		*/
		{
			s = spln( btfp_status[minor (mmdev)].level );	
									/* Block interrupts */

										
			if ( (usmask & MASK_DPA)				/* DPA already? */
				&& (dp->btfp_mask & (ushort)*arg) ) /* and user wants DPA */
			{
				if ( !(sp->gps_synch) )				/* Are we in synch with GPS? */
				{
					dp->btfp_ack = ACK_CLEAR;		/*  Clear OUTFIFO		*/
													/* When we get the next */
													/* DPA, the interrupt   */
													/* routine will get us  */
													/* synch-ed up with GPS */
													/* and set the flag     */
				}
				dp->btfp_ack = ACK_DPA;			/* Clear DPA status 		*/
			} /* end of "DPA-already" section */

			sp->intwait = (ushort)*arg;		/* Indicate we're waiting for an int.	*/
			dp->btfp_intstat = (ushort)*arg;/* Clear any existing interrupt */
									/* Save the interrupts we're waiting for */
			usmask = dp->btfp_intstat;				/* Clear ALL int. status */
			dp->btfp_intstat = usmask;				/* ... */
			dp->btfp_mask = (ushort)*arg; /* Enable bc63{57} interrupt(s).		*/

			while ( sp->intwait )   /* btfpintr() will clear this on int.	*/
			{						/* Block, waiting for a bc63{57} interrupt.*/
				sleep( (caddr_t)&sp->intwait, PZERO + 5 ); 
									/* Sleep till our interrupt occurs.		*/
			}
			splx( s );				/* Allow our device interrupts again.	*/
		}
		else						/* Signal sending IS enabled.			*/
		{
			s = spln( btfp_status[minor (mmdev)].level );	
									/* Block interrupts */
			if (usmask & MASK_DPA)			/* DPA already? */
			{
				if ( !(sp->gps_synch) )		/* Are we in synch with GPS? */
				{
					dp->btfp_ack = ACK_CLEAR;		/*  Clear OUTFIFO		*/
													/* When we get the next */
													/* DPA, the interrupt   */
													/* routine will get us  */
													/* synch-ed up with GPS */
													/* and set the flag     */
				}
				dp->btfp_ack = ACK_DPA;			/* Clear DPA status 		*/

			} /* end of "DPA-already" section */

			sp->intwait = (ushort)*arg;		/* We will not be blocking*/
			dp->btfp_intstat = (ushort)*arg;/* Clear any existing interrupt */
			usmask = dp->btfp_intstat;				/* Clear ALL int. status */
			dp->btfp_intstat = usmask;				/* ... */
			dp->btfp_mask = (ushort)*arg;/* Enable bc63{57} interrupt(s) etc.*/
			splx( s );				/* Allow interrupts from our device again.*/
									/* Return without blocking with a sleep.*/
		}
		break;
	
	case RINTSTAT:					/* Read the INTSTAT register.			*/
   		ustemp = dp->btfp_intstat;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		*arg   = uitemp;
		break;

	case WINTSTAT:						/* Write to the INTSTAT register.	*/
   		dp->btfp_intstat = (ushort)*arg;/* Directly from the user's argument.	*/
		break;
	
	case RVECTOR:					/* Read the VECTOR register.			*/
   		ustemp = dp->btfp_vector;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		*arg   = uitemp;
		break;
	
	case RLEVEL:					/* Read the LEVEL register.				*/
   		ustemp = dp->btfp_level;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		*arg   = uitemp;
		break;

	case READTIME:					/* Capture and read all time words.		*/
		ustemp		 		= dp->btfp_timereq;		/* Capture time.		*/
		tp = (struct btfp_time *)arg;
		tp->btfp_time[ 0 ] 	= dp->btfp_time[ 0 ];	/* Time word 0.			*/
		tp->btfp_time[ 1 ] 	= dp->btfp_time[ 1 ];	/* Time word 1.			*/
		tp->btfp_time[ 2 ] 	= dp->btfp_time[ 2 ];	/* Time word 2.			*/
		tp->btfp_time[ 3 ] 	= dp->btfp_time[ 3 ];	/* Time word 3.			*/
		tp->btfp_time[ 4 ] 	= dp->btfp_time[ 4 ];	/* Time word 4.			*/
		break;

	case READEVENT:					/* Read all event words.				*/
		ep = (struct btfp_event *)arg;
		ep->btfp_event[ 0 ] = dp->btfp_event[ 0 ];	/* Event word 0.		*/
		ep->btfp_event[ 1 ] = dp->btfp_event[ 1 ];	/* Event word 1.		*/
		ep->btfp_event[ 2 ] = dp->btfp_event[ 2 ];	/* Event word 2.		*/
		ep->btfp_event[ 3 ] = dp->btfp_event[ 3 ];	/* Event word 3.		*/
		ep->btfp_event[ 4 ] = dp->btfp_event[ 4 ];	/* Event word 4.		*/
		break;

	case WRITESTROBE:					/* Write to the Strobe registers.	*/
		dp->btfp_cr0 = STRDIS;			/* First disable strobe output		*/
		strp = (struct btfp_strobe *)arg;
		dp->btfp_strobe1 = strp->btfp_strobe[0];	/* Strobe word 1		*/
		dp->btfp_strobe2 = strp->btfp_strobe[1];	/* Strobe word 2		*/
		dp->btfp_strobe3 = strp->btfp_strobe[2];	/* Strobe word 3		*/
		break;

	case READALLREGS:				/* Read all bc63{57} regs.				*/
		up = (struct btfp_device *)arg;
		up->btfp_id		= dp->btfp_id;			/* ID register.				*/
		up->btfp_device = dp->btfp_device;		/* Device Type register.	*/
		up->btfp_status	= dp->btfp_status;		/* Status register.			*/

		up->btfp_timereq	= dp->btfp_timereq;		/* Capture the time.	*/
		up->btfp_time[ 0 ] 	= dp->btfp_time[ 0 ];/* Time word 0.			*/
		up->btfp_time[ 1 ] 	= dp->btfp_time[ 1 ];	/* Time word 1.			*/
		up->btfp_time[ 2 ] 	= dp->btfp_time[ 2 ];	/* Time word 2.			*/
		up->btfp_time[ 3 ] 	= dp->btfp_time[ 3 ];	/* Time word 3.			*/
		up->btfp_time[ 4 ] 	= dp->btfp_time[ 4 ];	/* Time word 4.			*/

		up->btfp_event[ 0 ] = dp->btfp_event[ 0 ];	/* Event word 0.		*/
		up->btfp_event[ 1 ] = dp->btfp_event[ 1 ];	/* Event word 1.		*/
		up->btfp_event[ 2 ] = dp->btfp_event[ 2 ];	/* Event word 2.		*/
		up->btfp_event[ 3 ] = dp->btfp_event[ 3 ];	/* Event word 3.		*/
		up->btfp_event[ 4 ] = dp->btfp_event[ 4 ];	/* Event word 4.		*/

		up->btfp_unlock 	= dp->btfp_unlock;		/* UNLOCK register.		*/
	
		up->btfp_ack 		= dp->btfp_ack;			/* ACK register.		*/
		up->btfp_cr0 		= dp->btfp_cr0;			/* Control register 0.	*/
		up->btfp_outfifo	= dp->btfp_outfifo;		/* OUTFIFO register.	*/
		up->btfp_mask		= dp->btfp_mask;		/* MASK register.		*/
		up->btfp_intstat	= dp->btfp_intstat;		/* INTSTAT register.	*/
		up->btfp_vector		= dp->btfp_vector;		/* VECTOR register.		*/
		up->btfp_level		= dp->btfp_level;		/* LEVEL register.		*/
	
		break;

	case SIGNALON:					/* Turn on signal sending on interrupts.*/
		sp->signalon = 1;			/* Indicate we should send signals on ints.*/
		break;
		
	case SIGNALOFF:					/* Turn off signal sending on interrupts.*/
		s = spln( btfp_status[minor (mmdev)].level );	/* Block interrupts */
		sp->signalon = 0;			/* Stop sending signals on interrupts.	*/
		sp->intwait = 0;			/* Indicate we are not awaiting any int.*/
		dp->btfp_mask = 0;			/* Turn off all 5 interrupts.			*/
		splx( s );					/* Allow interrupts from our device again.*/
		
		break;

	case PACKETON:					/* Turn on packet mode for read/write.*/
		sp->packeton = 1;			
		break;

	case PACKETOFF:					/* Turn off packet mode for read/write.*/
		sp->packeton = 0;			
		break;
		
	case DEBUGON:
		sp->debugon = 1;			/* Enable driver debugging 				*/
		break;
		
	case DEBUGOFF:
		sp->debugon = 0;			/* Disable driver debugging 			*/
		break;

	default:							
		u.u_error = EINVAL;			/* Return this is not a tty type device.*/
		break;
	}
	return( 0 );					/* Must return OK if no error found.	*/

} /* end of btfpioctl routine */


/************************************************************************/
/*                                                 
  	btfp_intr()- Called when the bc63{57} interrupts the processor.                                      
  								                           
  					This routine assumes that the interrupt level
  					on entry is the same as for the device, so that
  					we do not have to do an spl call.  If 
  					btfp_status[ unit].intwait does not match the interrupt
					which has just occurred, we were not
  					expecting this interrupt and we print an error message
  					and return after masking off module interrupt bits.  
  					Otherwise, we clear btfp_status[ unit].intwait,
  					clear the board interrupt bits, and wakeup the sleeping
  					ioctl WMASK routine, unless SIGNALON mode is in effect,
  					in which case we merely generate a signal and exit.              
                                                               
*/                                                                
/************************************************************************/


btfp_intr( my_isc, vector ) 
struct isc_table_type *my_isc;		/* pointer to isc structure 		*/
int vector;							/* Interrupt vector 				*/
{
	int i = 0;
	int mmdev;
	register BTFPVOLATILE *dp;
	register struct btfp_status *sp;
	unsigned short 	ustemp, ustemp2; /* Temps for register access 		*/
	unsigned short 	usmask = 
					MASK_INT0 | MASK_INT1 |MASK_INT2 | MASK_INT3 | MASK_INT4;
									/* Mask of module interrupt enable bits.*/
	i = 0;
	while( i <= MAXDEV )
	{
		if(btfp_status[i].vector == (vector & 0xff)) break; 
		i++;
	}
	if ( i > MAXDEV)
	{
		printf("btfp_intr:Undefined interrupt vector received!\n");
		return;
	}
	else
	{
		mmdev = i;
	}	

	dp = btfp_addr[ mmdev ];
   										/* Init pointer to base register.	*/
	sp = &btfp_status[ mmdev ];
										/* Init pointer to status structure.*/

	ustemp = dp->btfp_intstat;			/* Read INTSTAT register.			*/
	
	if( !(sp->intwait & ustemp) )		/* We were not waiting for this int!*/
	{
		dp->btfp_mask = 0;			 	/* Turn off all 5 interrupts.		*/
		dp->btfp_intstat =  usmask; 	/* Turn off all 5 interrupts.		*/
		printf("btfpintr(): Unexpected interrupt from btfp%d\n", minor(mmdev));
		return;							/* Return to caller.				*/                               
	}
	else if( !sp->signalon )			/* We were waiting for an int..		*/
	{									/* and we were blocked doing so.	*/
		sp->intwait = 0;				/* Indicate we received an interr.	*/
		if (ustemp & INTSTAT_DPA)
		{
			/* For DPA interrupt, remember to clear the ACK_DPA bit */
			dp->btfp_ack = ACK_DPA;
		}

		dp->btfp_mask = 0;				/* Turn off all 5 interrupts.		*/
		dp->btfp_intstat = usmask; 		/* Turn off any interrupts.			*/
		wakeup( (caddr_t)&sp->intwait );/* Allow the ioctl WMASK to proceed.*/
	}
	else								/* Waiting for an interrupt but in 	*/
	{									/* SIGNALON mode...					*/
		dp->btfp_intstat = usmask; 		/* Turn off any interrupts.			*/

		/* If DPA is here... */
		if (ustemp & INTSTAT_DPA) 
		{

		if (sp->debugon)
		{
			if (ustemp & INTSTAT_DPA)
				printf ("d");
		}

		/* ...AND the user wanted to look at it */
 		if (sp->intwait & MASK_DPA) 
		{
		if (sp->debugon)
		{
			if (ustemp & INTSTAT_DPA)
				printf ("D");
		}

		dp->btfp_ack = ACK_DPA;				/* Clear DPA bit in ACK reg 	*/

		/* NOTE: SIGNALON mode has already waited for the DPA signal, 		*/
		/* thus its read just reads the available packet 					*/

		/* Now fill packet buffer from OUTFIFO */

		sp->rpacki = 0;

		/* We must read the first byte to activate the ACK_MORE flag 		*/
		/* and to see if we are out of synch with the TANS packets.  		*/
		/* If the first two bytes don't look like the start of a     		*/
		/* packet, just clear the OUTFIFO again.					 		*/

		if (!sp->gps_synch)
		{

		sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;	/* Get first byte */
		if ( sp->rpack_buffer[sp->rpacki - 1] != (unsigned char) DLE )
		{
			/* Bogus packet...clear OUTFIFO */
			dp->btfp_ack = ACK_CLEAR;		/* Clear OUTFIFO 				*/
		}
		else
		{
			/* First byte OK...now look at next byte: is it a packet ID? */
			sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;	/* Get next byte */
			if  ( ( sp->rpack_buffer[sp->rpacki - 1] == (unsigned char) DLE )
				|| ( sp->rpack_buffer[sp->rpacki - 1] == (unsigned char) ETX ) )
			{
				/* Bogus packet...clear OUTFIFO */
				dp->btfp_ack = ACK_CLEAR;		/* Clear OUTFIFO */
			}
			/* else packet is OK...read it in */
				
		} /* end of first byte OK section */

		if (sp->debugon)
			printf ("S");
		sp->gps_synch = 1;						/* We're synched up now 	*/

		} /* end of "need to synch up with the GPS" section */

		/* NOTE: ACK_MORE bit on means "more bytes in OUTFIFO" */

		ustemp = dp->btfp_ack;
		while ( (ustemp & ACK_MORE) 		/* Until FIFO is empty... 		*/
			&& (sp->rpacki < MAXPACKET) )	/* or running away 				*/
		{
			sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;	/* Get next byte */
			ustemp = dp->btfp_ack;			/* Any more? 					*/
			
		} /* end of OUTFIFO read loop */
		
		if (sp->rpacki == 0)
		{
			if (sp->debugon)
				printf ("0");
			/* Wait for next DPA interrupt */
			return;
		}

		} /* end of "user wanted to see DPA" section */

		else
		{
			ustemp &= ~INTSTAT_DPA;			/* Any other interrupts happen?	*/
			if ( !(ustemp & sp->intwait) )	/* If not our interest,bail out */
			{
				return;
			}
			/* else, fall through to send the user a signal */
		}

		} /* end of DPA section */

		/* if no DPA, we must have matched on another INTSTAT bit */

		psignal( sp->procp, SIGUSR1 );		/* Send signal to the application.*/
	} /* end of SIGNALON section */

} /* end of btfpintr routine */

static drv_ops_t btfp_ops = {
        btfp_open,
        btfp_close,
        NULL,
        NULL,
        NULL,
        NULL,
        btfp_read,
        btfp_write,
        btfp_ioctl,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL
};
static drv_info_t btfp_info = {
        "btfp",
        "unknown",
        DRV_CHAR | DRV_SAVE_CONF,
        -1,
        -1,
        NULL,
        NULL,
        NULL
};
static wsio_drv_data_t btfp_data = {
        "btfp",
        T_DEVICE,
        DRV_CONVERGED,
        NULL,
        NULL
};
static wsio_drv_info_t btfp_wsio_info = {
        &btfp_info,
        &btfp_ops,
        &btfp_data
};



/****************************  INSTALL  ******************************
 * Registers the driver attach function and registers the WSIO data
 * structures that specify the function and configuration data
 * associated with the driver.
 *********************************************************************/
btfp_install()
{
        vme_set_attach_function(BTFP_STRING, btfp_attach);
        return(wsio_install_driver(&btfp_wsio_info));
}


