#ident "@(#)btfptest.c	1.7"

/************************************************************************/
/*                                                                      
  	BANCOMM bc63{57}VME/ bc350VXI Time and Frequency Processor 
  			Test Application for HP-UX v9.01 Systems.                
 	
 	Copyright (C) 1992, BANCOMM, San Jose, CA.  All rights reserved.	
                                                                     
  	Author:	Greg Dowd	
*/                                                                      
/************************************************************************/
/*
	Description:  This program is used to test the bc63{57} driver functions.
					It is not specifically designed to test
					the bc63{57} module.
*/
/************************************************************************/
/*
	Command line options select various tests to be run, as well as other
		test options. Options are processed left to right, so the
		rightmost option takes precedence.

	btfptest -vSaALRTEDrilp -b <block test #> -s <signal test #> -d <display #>

	where:
	-a : Run all tests, excluding tests g, b1, and s1.
	-A : Run ALL tests (assumes external interrupt generator and GPS attached)
	-v : verbose mode (prints diagnostic displays to stdout)
	-S : Slow terminal (defaults to fast terminal)
	-L : Loop through all selected tests, as appropriate (default: no loop)
	-R : READALLREGS test. Reads all bc63{57} registers.
	-T : READTIME test.
	-E : READEVENT test.
	-I : Interactive mode (affects heartbeat/external-interrupt tests only)
	-D : enable driver debugging
	-r : read(2)/write(2) test (default mode).
	-i : enable simple ioctl(2) tests 
		(excludes READTIME, READEVENT, READALLREGS)
	-l : load major time test. Uses current time from system clock.
	-g : select GPS test. Writes to INFIFO, reads from OUTFIFO.
	-b : select block-until-interrupt test
		0 : no block-until-interrupt tests
		1 : external interrupt test
		2 : heartbeat test
		3 : major/minor strobe test (one-time)
		4 : minor-only strobe test (once per second)
		5 : 1PPS test
		6 : all block-until-interrupt tests
	-s : select signal handling test
		0 : no signal handling tests
		1 : external interrupt test
		2 : heartbeat test
		3 : major/minor strobe test (one-time)
		4 : minor-only strobe test (once per second)
		5 : 1PPS test
		6 : all signal handling tests
	-d : select display mode when heartbeat signal received
		0 : no heartbeat status display
		1 : display time of heartbeats (DEFAULT)
		2 : display count of heartbeats
		3 : display time and count of heartbeats
*/
/************************************************************************/
/*	Headers containing definitions used in this program.
*/
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>


#include "btfpu.hpux.h"

#define	BLOCK	1	/* Blocking mode */
#define	SIGNAL	2	/* Signaling mode */

extern void signalhandler();
extern void hbsighandler();
extern void wtsighandler();
extern void rpsighandler();
extern char *optarg;
/************************************************************************/
struct btfp_device ds;				/* Device structure for use with 	*/
									/* READALL							*/
struct btfp_time ts;					/* 16bit time words. */
									/* Used with READTIME ioctl call.	*/
struct btfp_time t2s;				/* Same thing for the machine speed test.*/
struct btfp_event event;				/* Event words */
									/* Used with READEVENT ioctl call.	*/
struct btfp_strobe strobe;				/* Strobe words */
									/* Used with WRITESTROBE ioctl call.	*/

int signalcount;					/* Used for counting received signals.*/
int	printcount;						/* Limits printing of signal count.	*/
int	prints_done;					/* Limits looping of heartbeats.	*/
long per_second;					/* Heartbeats per second			*/
int bytes_read;						/* Packet bytes read				*/

/************************************************************************/
    int fildes;                             /* Return value from open.      */
    char* bc635name = "/dev/btfp0";          /* Name of the bc63{57} device.   */
    char    buf[ 32 ];                      /* Place to put time read in.   */
    char    time_buf[ 12 ];                  /* Place to put jjjHHMMSS date. */
	int		num_chars;						/* Number of chars in time_buf */
    unsigned    nbytes;                     /* Number of bytes to read.     */
    unsigned    numread;                    /* # bytes actually read.       */
    int         regvalue, newvalue;         /* Register value returned.     */
    int         t;                          /* Temporary time holder.       */
    int         time0;                      /* Time values for general use. */
    int         time1;
    int         time2;
    int         time3;
    int         i;                          /* Loop index thanks to K&R.    */
    unsigned int    uitemp;                 /* Temporary unsigned integer.  */
    unsigned int    uihold;                 /* Temporary unsigned integer.  */
    time_t          unixtime;               /* Returned by 'time'.          */
    int		err;
	unsigned char wpacket[MAXPACKET];				/* for write to INFIFO */
	unsigned char rpacket[MAXPACKET];				/* for read from INFIFO */
	char delimiters[2]={' ','\n'};				/* for build_packet */
	int time_requested;							/* GPS test globals */
	int time_received;
	int setup_done;
	unsigned char packet_21[7]={			/* "Request Current Time" packet */
		SOH, 'J', ETB, DLE, 0x21, DLE, ETX}; /* for GPS Satellite Receiver */

    void usage(),
		do_readall_test(),
		do_readtime_test(),
		do_readevent_test(),
		do_read_test(),
		do_ioctl_test(),
		do_loadtime_test(),
		do_GPS_test(),
		do_int_tests ();
	void set_heartbeat(),
		set_strobe(),
		write_time(),
		send_packet(),
		write_packet(),
		read_packet(),
		build_packet(),
		disp_packet();

    /* Set up default flags... */
    int	verbose = NO;
    int	fast_terminal = YES;
    int	all_tests = NO;
    int	ALL_tests = 0;
    int	test_picked = NO;
    int	loop_tests = NO;
    int	readall_test = NO; /* READALLREGS */
    int	readtime_test = NO; /* READTIME */
    int	readevent_test = NO; /* READEVENT */
    int	read_test = NO; /* read(2) */
    int	ioctl_tests = NO;
    int	loadtime_test = NO;

    /* ...and default modes */
    int	GPS_test = 0;
    int	block_tests = 0;
    int	signal_tests = 0;
    int	display_mode = 1;					/* Default to display time */
    int	interact_mode = 0;

/************************************************************************/
main (argc, argv)
int argc;
char **argv;
{
	int dummy;

	printf( "\n\n\n" );
	printf( "BANCOMM bc63{57}VME/bc350VXI Time and Frequency Processor, Driver Test Program.\n" );
	time( &unixtime );						/* Get current 'unix-format' time.*/
	printf( "\n %s", ctime( &unixtime ) );	/* Log out the current date/time.*/
	
	/* Check arguments */
	if (argc == 1) 
	{
		usage(argv[0]);
		exit (-1);
	}
	
/****************************************************************************/
	printf( "\n\n Open(2) test (standard UNIX open of the device)...\n" );
	
	fildes = open( bc635name, O_RDWR );	
											/* Open the bc63{57} device. 		*/
	if( fildes == -1 )
	{
		printf( "   UNABLE TO OPEN bc63{57} DEVICE %s!\n\n", bc635name );
		exit( 0 );							/* This ends the test.			*/
	}
	else
	{
		printf( "   Time and frequency processor %s opened successfully.\n", bc635name );
		if( ioctl( fildes, RUNLOCK, &dummy ) ) pioerr(__LINE__);	
											/* Release capture lockout in 	*/
											/* case it was set from before.	*/
		regvalue = 0;
		if( ioctl( fildes, WCR0, &regvalue ) ) pioerr(__LINE__); /* Clean up from kill on	*/
	} /* end of successful open case */
	
/****************************************************************************/

	while ((err = getopt (argc, argv, "vSaALRTEIDrilgb:s:h:d:"))
			!= -1)
		switch (err) 
		{
		case 'v':
			verbose = YES;
			printf ("Enabling verbose mode\n");
			break;

		case 'S':
			if (verbose)
				printf ("Slow Terminal Mode\n");
			fast_terminal = NO;
			break;

		case 'a':
			if (verbose)
				printf ("All Tests Mode (except for interrupt tests)\n");
			all_tests = YES;
			test_picked = YES;
			readall_test = YES; /* READALLREGS */
			readtime_test = YES; /* READTIME */
			readevent_test = YES; /* READEVENT */
			read_test = YES; /* read(2) */
			ioctl_tests = YES;
			loadtime_test = YES;
			block_tests = 6;	/* Won't do external interrupt test */
			signal_tests = 6;	/* Won't do external interrupt test */
			break;

		case 'A':
			if (verbose)
				printf ("ALL TESTS Mode (including external tests)\n");
			ALL_tests = YES;
			test_picked = YES;
			readall_test = YES; /* READALLREGS */
			readtime_test = YES; /* READTIME */
			readevent_test = YES; /* READEVENT */
			read_test = YES; /* read(2) */
			ioctl_tests = YES;
			loadtime_test = YES;
			GPS_test = YES;
			block_tests = 6;
			signal_tests = 6;
			/* Note that display option still defaults to OFF */
			break;

		case 'L':
			if (verbose)
				printf ("Infinite Test-Loop Mode\n");
			loop_tests = YES;
			break;

		case 'R':
			if (verbose)
				printf ("Read All Registers Test\n");
			readall_test = YES;
			test_picked = YES;
			break;

		case 'T':
			if (verbose)
				printf ("READTIME Test\n");
			readtime_test = YES;
			test_picked = YES;
			break;

		case 'E':
			if (verbose)
				printf ("READEVENT Test\n");
			readevent_test = YES;
			test_picked = YES;
			break;

		case 'I':
			interact_mode = YES;
			break;

		case 'D':
			if (verbose)
				printf ("Enable Driver debugging\n");
			regvalue = 0;
			if( ioctl( fildes, DEBUGON, &regvalue ) ) pioerr(__LINE__); /* Set driver debug flag */
			break;

		case 'r':
			if (verbose)
				printf ("read(2) Test\n");
			read_test = YES;
			test_picked = YES;
			break;

		case 'i':
			if (verbose)
				printf ("ioctl(2) Test\n");
			ioctl_tests = YES;
			test_picked = YES;
			break;

		case 'l':
			if (verbose)
				printf ("Load Major Time Test\n");
			loadtime_test = YES;
			test_picked = YES;
			break;

		case 'g':
			if (verbose)
				printf ("GPS Test\n");
			GPS_test = YES;
			test_picked = YES;
			break;

		case 'b':
			block_tests = atoi (optarg);
			switch (block_tests)
			{
				case 0:
					if (verbose)
						printf ("No blocking tests\n");
					break;
				case 1:
					if (verbose)
						printf ("External Interrupt blocking test\n");
					break;
				case 2:
					if (verbose)
						printf ("Heartbeat blocking test\n");
					break;
				case 3:
					if (verbose)
						printf ("Major/Minor Strobe blocking test\n");
					break;
				case 4:
					if (verbose)
						printf ("Minor-only Strobe blocking test\n");
					break;
				case 5:
					if (verbose)
						printf ("1PPS blocking test\n");
					break;
				case 6:
					if (verbose)
						printf ("ALL blocking tests\n");
					break;
				default:
					fprintf (stderr, "Invalid Block Test Option: %d\n",
						block_tests);
					exit (1);
			} /* end of "switch (optarg)" */
			test_picked = YES;
			break;

		case 's':
			signal_tests = atoi (optarg);
			switch (signal_tests)
			{
				case 0:
					if (verbose)
						printf ("No signaling tests\n");
					break;
				case 1:
					if (verbose)
						printf ("External Interrupt signaling test\n");
					break;
				case 2:
					if (verbose)
						printf ("Heartbeat signaling test\n");
					break;
				case 3:
					if (verbose)
						printf ("Major/minor Strobe signaling test\n");
					break;
				case 4:
					if (verbose)
						printf ("Minor-only Strobe signaling test\n");
					break;
				case 5:
					if (verbose)
						printf ("1PPS signaling test\n");
					break;
				case 6:
					if (verbose)
						printf ("ALL signaling tests\n");
					break;
				default:
					fprintf (stderr, "Invalid Signal Test Option: %d\n",
						signal_tests);
					exit (1);
			} /* end of "switch (optarg)" */
			test_picked = YES;
			break;

		case 'd':
			display_mode = atoi (optarg);
			switch (display_mode)
			{
				case 0:
					if (verbose)
						printf ("No display for heartbeat tests\n");
					break;
				case 1:
					if (verbose)
						printf ("Display time for heartbeat tests\n");
					break;
				case 2:
					if (verbose)
						printf ("Display count for heartbeat tests\n");
					break;
				case 3:
					if (verbose)
						printf ("Display time and count for heartbeat tests\n");
					break;
				default:
					fprintf (stderr, "Invalid Display Mode: %d\n",
						display_mode);
					exit (1);
			} /* end of "switch (optarg)" */
			break;

		default:
				fprintf (stderr, "Invalid option\n");
		case '?':
				(void) usage (argv[0]);
				exit (-1);

		} /* end of "switch (err)" */

		if (test_picked == NO)
		{
			printf ("No test selected.\n");
			exit (1);
		}
	
/****************************************************************************/
/* Put board into Mode 1 for standard driver testing						*/

	send_packet("A1");
/****************************************************************************/

		/* Main section of test */

	do 
	{

		if (readall_test)
			(void) do_readall_test();

		if (readtime_test)
			(void) do_readtime_test();

		if (readevent_test)
			(void) do_readevent_test();

		if (read_test)
			(void) do_read_test();

		if (ioctl_tests)
			(void) do_ioctl_test();

		if (loadtime_test)
			(void) do_loadtime_test();

		if (GPS_test)
			(void) do_GPS_test();

		if (block_tests)
			(void) do_int_tests (block_tests, BLOCK);

		if (signal_tests)
			(void) do_int_tests (signal_tests, SIGNAL);

	} /* end of Big Loop */
	while (loop_tests == YES);

	if (all_tests == YES)
		printf ("\nAll standalone tests completed!\n");

	if (ALL_tests == YES)
		printf ("\nAll tests completed!\n");

} /* end of main routine */
/****************************************************************************/
void
do_readall_test()
{
	
	printf( "\n   READALLREGS- Read all readable bc63{57} registers at once...\n\n" );
	if( ioctl( fildes, READALLREGS, &ds ) ) pioerr(__LINE__);
	printregs();							/* Print out what they all are.	*/

} /* end of do_readall_test routine */
/****************************************************************************/
void
do_readtime_test()
{

	printf( "\nREADTIME ioctl call test (read time in one call)...\n" );
	printf( "     Two back-to-back such reads indicate call timing:\n\n" );
	
	if( ioctl( fildes, READTIME, &ts ) ) pioerr(__LINE__);
												/* Read all 5 time registers.*/
	if( ioctl( fildes, READTIME, &t2s ) ) pioerr(__LINE__);
	
	for (i=0; i < 5; i++)
	{
		printf( "      READTIME- Time words %d: 0x%04x   0x%04x\n",
							 i, ts.btfp_time[ i ] , t2s.btfp_time[ i ]);
	}

	printf ("READTIME test completed\n");

} /* end of do_readtime_test routine */
/****************************************************************************/
void
do_readevent_test()
{
	int dummy;

	/* READEVENT test */

	printf ("\nREADEVENT test\n");

	dummy = 0;
	/* first disable interrupts */
	if( ioctl( fildes, WMASK, &dummy ) ) pioerr(__LINE__); 

	/* clear any existing interrupt status bits */
	if( ioctl( fildes, RINTSTAT, &regvalue ) ) pioerr(__LINE__); 
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	/* clear any time capture lockout */
	if( ioctl( fildes, RUNLOCK, &regvalue ) ) pioerr(__LINE__); 

	regvalue = LOCKEN | HBEN;
	/* enable heartbeat time capture, and time capture lockout */
	if( ioctl( fildes, WCR0, &regvalue ) ) pioerr(__LINE__); 

	/* set up the heartbeat signal to be 1 count per second, */
	/* synchronized with the 1 PPS */

	set_heartbeat( 10000, 1000);
	
	regvalue = INTSTAT_HEART;
	/* clear heartbeat interrupt */
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	/* wait for heartbeat to occur */
	regvalue = 0;
	do
	{
		if( ioctl( fildes, RINTSTAT, &regvalue ) ) pioerr(__LINE__); 
	}
	while (!(regvalue & INTSTAT_HEART));

	dummy = INTSTAT_HEART;
	/* clear the heartbeat status */
	if( ioctl( fildes, WINTSTAT, &dummy ) ) pioerr(__LINE__); 

	/* Release capture lockout. */
	if( ioctl( fildes, RUNLOCK, &dummy  ) ) pioerr(__LINE__); 

	/* Event registers should now have the captured time */
	get_event();

	dummy = 0;
	/* Clear the CR0 register */
	if( ioctl( fildes, WCR0, &dummy ) ) pioerr(__LINE__); 

	printf ("READEVENT test completed\n");
} /* end of do_readevent_test routine */

/****************************************************************************/
void
do_read_test()
{
	int rtn;
	time_t timeval;

	/* Note: this routine exercises the "default" read/write mode, */
	/* which accesses the time in Free Running Mode (mode 1) */
	/* This mode corresponds to PACKETOFF */

	printf( "\n\n Read(2)/Write(2) test (default mode)...\n" );

	regvalue = 0;
	/* Ensure that we are in default mode for reads and writes */
	if( ioctl( fildes, PACKETOFF, &regvalue ) ) pioerr(__LINE__); 
	
	printf ("\t3 Time reads before SIGNALOFF write call...\n");

	for( i = 0; i < 3; i++ )				/* Read the time and display it.*/
	{
		sleep( 1 );							/* Wait a second and go again.*/
		nbytes = 32;							/* Ask for all 32 bytes.	*/
		numread = read( fildes, buf, nbytes );/* Read time from bc63{57} device.*/
		if( numread == 0 ) 						/* Then we are at 'end of file'.*/
		{
			lseek( fildes, 0L, 0 );				/* 'Rewind' the device.		*/
			numread = read( fildes, buf, nbytes );	/* Read time from bc63{57}. */
		}
		else if( numread == -1 )
		{
			printf( "      ERROR READING TIME FROM %s!\n",  bc635name );
		}
		printf( "      Time read is: %s\n", buf );	/* Show the time read.	*/
	} /* end of for loop */

	/* Now write to the BTFP real time clock from the PC's system clock */
	/* First in SIGNALOFF mode */
	write_time (SIGNALOFF);

	if (verbose)
		printf ("Wrote current time from PC system clock to BTFP (SIGNALOFF mode)\n");
	printf ("\t3 Time reads after SIGNALOFF write call and before SIGNALON write call...\n");

	/* Now read the time again */
	
	for( i = 0; i < 3; i++ )				/* Read the time and display it.*/
	{
		sleep( 1 );							/* Wait a second and go again.*/
		nbytes = 32;							/* Ask for all 32 bytes.	*/
		numread = read( fildes, buf, nbytes );	/* Read time from bc63{57} device.*/
		if( numread == 0 ) 						/* Then we are at 'end of file'.*/
		{
			lseek( fildes, 0L, 0 );				/* 'Rewind' the device.		*/
			numread = read( fildes, buf, nbytes );	/* Read time from bc63{57}. */
		}
		else if( numread == -1 )
		{
			printf( "      ERROR READING TIME FROM %s!\n",  bc635name );
		}
		printf( "      Time read is: %s\n", buf );	/* Show the time read.	*/
	} /* end of for loop */

	/*  Write again to the BTFP real time clock from the PC's system clock */

	/* Next in SIGNALON mode */
	write_time (SIGNALON);

	if (verbose)
		printf ("Wrote current time from PC system clock to BTFP (SIGNALON mode)\n");
	printf ("\t3 Time reads after SIGNALON write call...\n");
	/* Now read the time again */
	
	for( i = 0; i < 3; i++ )				/* Read the time and display it.*/
	{
		sleep( 1 );							/* Wait a second and go again.*/
		nbytes = 32;							/* Ask for all 32 bytes.	*/
		numread = read( fildes, buf, nbytes );	/* Read time from bc63{57} device.*/
		if( numread == 0 ) 						/* Then we are at 'end of file'.*/
		{
			lseek( fildes, 0L, 0 );				/* 'Rewind' the device.		*/
			numread = read( fildes, buf, nbytes );	/* Read time from bc63{57}. */
		}
		else if( numread == -1 )
		{
			printf( "      ERROR READING TIME FROM %s!\n",  bc635name );
		}
		printf( "      Time read is: %s\n", buf );	/* Show the time read.	*/
	} /* end of for loop */

} /* end of do_read_test routine */

/****************************************************************************/
void
do_ioctl_test()
{
	printf( "\n\n Ioctl(2) call tests...\n" );

	/* Configuration Registers */

	printf( "\n\tRIDR - Read from the ID register...\n");
	if( ioctl( fildes, RIDR, &regvalue ) ) pioerr(__LINE__);	/* Read ID register */

	printf ( "\tRIDR - ID register is 0x%04x.\n\n", regvalue);

	printf( "\n\tRDTR - Read from the DT register...\n");
	
	if( ioctl( fildes, RDTR, &regvalue ) ) pioerr(__LINE__);	/* Read DT register */

	printf ( "\tRDTR - DT register is 0x%04x.\n\n", regvalue);

	printf( "\n\tRSR - Read from the Status register...\n");
	
	if( ioctl( fildes, RSR, &regvalue ) ) pioerr(__LINE__);	/* Read Status register */

	printf ( "\tRSR - Status register is 0x%04x.\n\n", regvalue);

	/* General Registers */

	/* Write to Control Register here? */

	/* NOTE: this would then require setting up LEVEL/VECTOR, which are */
	/* currently set up only by the driver */
	
	/* Time Request Registers */

	if( ioctl( fildes, RTIMEREQ, &regvalue ) ) pioerr(__LINE__);	/* Read the time req reg.*/
											/* Ignore the read value */

	
	if( ioctl( fildes, RTIME0, &regvalue ) ) pioerr(__LINE__);	/* Read the 1st time reg.*/
	printf( "      RTIME0- Time word 0 is 0x%04x.\n", regvalue );
	if( ioctl( fildes, RTIME1, &regvalue ) ) pioerr(__LINE__);	/* Read the 2nd time reg.*/
	printf( "      RTIME1- Time word 1 is 0x%04x.\n", regvalue );								
	if( ioctl( fildes, RTIME2, &regvalue ) ) pioerr(__LINE__);	/* Read the 3rd time reg.*/
	printf( "      RTIME2- Time word 2 is 0x%04x.\n", regvalue );								
	if( ioctl( fildes, RTIME3, &regvalue ) ) pioerr(__LINE__);	/* Read the 4th time reg.*/
	printf( "      RTIME3- Time word 3 is 0x%04x.\n", regvalue );

	if( ioctl( fildes, RTIME4, &regvalue ) ) pioerr(__LINE__);	/* Read the 5th time reg.*/
	printf( "      RTIME4- Time word 4 is 0x%04x.\n", regvalue );
	
	if( ioctl( fildes, REVENT0, &regvalue ) ) pioerr(__LINE__);	/* Read the 1st event reg.*/
	/* Event Registers */

	printf ("\n\nNOTE: The content of these registers are meaningless for this test:\n");
	printf( "      REVENT0- Event word 0 is 0x%04x.\n", regvalue );
	if( ioctl( fildes, REVENT1, &regvalue ) ) pioerr(__LINE__);	/* Read the 2nd event reg.*/
	printf( "      REVENT1- Event word 1 is 0x%04x.\n", regvalue );								
	if( ioctl( fildes, REVENT2, &regvalue ) ) pioerr(__LINE__);	/* Read the 3rd event reg.*/
	printf( "      REVENT2- Event word 2 is 0x%04x.\n", regvalue );								
	if( ioctl( fildes, REVENT3, &regvalue ) ) pioerr(__LINE__);	/* Read the 4th event reg.*/
	printf( "      REVENT3- Event word 3 is 0x%04x.\n", regvalue );

	if( ioctl( fildes, REVENT4, &regvalue ) ) pioerr(__LINE__);	/* Read the 5th event reg.*/
	printf( "      REVENT4- Event word 4 is 0x%04x.\n", regvalue );

	/* Strobe Registers */

	printf ("\tWriting 0x66 to each Strobe register individually.\n");
	regvalue = 0x66;
	if( ioctl( fildes, WSTROBE1, &regvalue ) ) pioerr(__LINE__);/* Write the 1st strobe reg.*/
	if( ioctl( fildes, WSTROBE2, &regvalue ) ) pioerr(__LINE__);/* Write the 2nd strobe reg.*/
	if( ioctl( fildes, WSTROBE3, &regvalue ) ) pioerr(__LINE__);/* Write the 3rd strobe reg.*/
	printf ("\n   Strobe register writes OK \n");

/****************************************************************************/
	/* UNLOCK, ACK registers */

	printf ("\n\tReading from UNLOCK register.\n");
	if( ioctl( fildes, RUNLOCK, &regvalue ) ) pioerr(__LINE__);	/* Read the UNLOCK reg. */
										/* Ignore value read */
	printf ("\tUNLOCK register read OK \n");

	if( ioctl( fildes, RACK, &regvalue ) ) pioerr(__LINE__);	/* Read the ACK reg. */
	printf( " \n\tACK register is 0x%04x.\n", regvalue );

	/* write the same value back */
	if( ioctl( fildes, WACK, &regvalue ) ) pioerr(__LINE__);	/* Write the ACK reg. */

	printf ("\tWriting the same value back to the ACK register...\n");
	if( ioctl( fildes, RACK, &regvalue ) ) pioerr(__LINE__);	/* Read the ACK reg. */
	printf( " \n\tACK register is 0x%04x.\n", regvalue );
	printf ("\tACK register read/write OK \n");

/****************************************************************************/
	printf( "\n\tWCR0 - RCR0- Write/read control register 0...\n" );
	
	printf( "      RCR0- Read the register's current state...\n" );
	if( ioctl( fildes, RCR0, &regvalue ) ) pioerr(__LINE__);	
	printf(   "      RCR0- Control register 0 is 0x%04x.\n", regvalue );
	uihold = regvalue;						/* Save original state.	*/

	uitemp = 0x00;
	printf( "\n      WCR0- Write all zeros to the register...\n" );
	if( ioctl( fildes, WCR0, &uitemp ) ) pioerr(__LINE__); 	
	if( ioctl( fildes, RCR0, &regvalue ) ) pioerr(__LINE__);	
	printf(   "      RCR0- Control register 0 is 0x%04x.\n", regvalue );

	uitemp = 0x04;							/* Set a new value.		*/
	printf( "\n      WCR0- Write 0x%x to the register...\n", uitemp );
	if( ioctl( fildes, WCR0, &uitemp ) ) pioerr(__LINE__); 
	if( ioctl( fildes, RCR0, &regvalue ) ) pioerr(__LINE__);	
	printf(   "      RCR0- Control register 0 is 0x%04x.\n", regvalue );
									
	printf( "\n      WCR0- Clear it to all zeros, its hardware initialized state.\n" );
	regvalue = 0;
	if( ioctl( fildes, WCR0, &regvalue ) ) pioerr(__LINE__); 

	printf( "\n\tWCR0 - RCR0 Write/read control register 0 tests complete\n");
/****************************************************************************/
	/* FIFO basic tests; the GPS test actually exchanges real data */

	printf ("\n\tIN/OUT FIFO tests\n");
	printf ("\n\t\tNOTE: data are meaningless for this test\n");

	printf ("\t\tWriting zero to the INFIFO...\n");
	regvalue = 0; /* init to zero */
	if( ioctl( fildes, WINFIFO, &regvalue ) ) pioerr(__LINE__); 
	if( ioctl( fildes, ROUTFIFO, &regvalue ) ) pioerr(__LINE__); 
	printf ("\tRead 0x%0.2x from the OUTFIFO...\n", regvalue);
	printf ("\tIN/OUT FIFO tests completed\n");

/****************************************************************************/
	/* MASK register */

	printf ("\nWrite/Read MASK register tests\n");

	if( ioctl( fildes, RMASK, &regvalue ) ) pioerr(__LINE__); 
	/* Note: only bits 0-4 are used */
	printf(   "      MASK register is now  0x%04x.\n", regvalue & 0x1f);

	printf(   "      Clearing MASK register\n");

	regvalue = 0; /* init to zero */

	if( ioctl( fildes, WMASK, &regvalue ) ) pioerr(__LINE__); 
	if( ioctl( fildes, RMASK, &newvalue ) ) pioerr(__LINE__); 
	/* Note: only bits 0-4 are used */
	if (regvalue != (newvalue & 0x1f))
	{
		printf (" Error: MASK register mismatch (old/new): 0x%04x/0x%04x\n",
			regvalue, (newvalue & 0x1f));
	}
	else
	{
		printf(   "      MASK register set to  0x%04x.\n", (newvalue & 0x1f) );
	}

	printf ("MASK tests completed\n");

/****************************************************************************/
	/* INTSTAT register test */

	printf ("\nINTSTAT register tests\n");

	/* clear any existing interrupt status bits */
	if( ioctl( fildes, RINTSTAT, &regvalue ) ) pioerr(__LINE__); 
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	printf ("\tWaiting for the 1PPS status bit...\n");

	/* now look for the 1 PPS status bit */
	regvalue = 0;
	do {
		if( ioctl( fildes, RINTSTAT, &regvalue ) ) pioerr(__LINE__); 
		if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 
	}
	while ((regvalue & INTSTAT_1PPS) == 0);

	printf ("\tGot the 1PPS status bit.\n");

	/* Clear the 1PPS interrupt status */
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	regvalue = ACK_1PPS;
	/* clear the 1PPS ACK bit */
	if( ioctl( fildes, WACK, &regvalue ) ) pioerr(__LINE__);

	printf ("INTSTAT tests completed\n");

/****************************************************************************/
	/* VECTOR/LEVEL registers */

	printf ("\nVECTOR/LEVEL register tests\n");
	if( ioctl( fildes, RVECTOR, &regvalue ) ) pioerr(__LINE__); 
	printf("\tVECTOR register contents:  0x%04x.\n", regvalue );
	if( ioctl( fildes, RLEVEL, &regvalue ) ) pioerr(__LINE__); 
	printf("\tLEVEL register contents:  0x%04x.\n", regvalue );
	printf ("VECTOR/LEVEL tests completed\n");

/****************************************************************************/
	/* READTIME done separately */
/****************************************************************************/
	/* READEVENT done separately */

/****************************************************************************/
	/* WRITESTROBE test */

	printf ("\nWRITESTROBE test\n");

	/* just clear to zeroes */
	for (i=0; i < 5; i++)
	{
		strobe.btfp_strobe [i] = 0;
	}
	printf ("\tWriting zeros to all the strobe registers.\n");
	if( ioctl( fildes, WRITESTROBE, &strobe ) ) pioerr(__LINE__); 
	printf ("WRITESTROBE test completed\n");
/****************************************************************************/
	/* READALLREGS done separately */
/****************************************************************************/
	/* SIGNALON/OFF tests */
	printf ("\nSIGNALON/OFF test\n");
	printf ("\tSetting Signal Handling Mode (SIGNALON).\n");
	regvalue = 0;
	if( ioctl( fildes, SIGNALON, &regvalue ) ) pioerr(__LINE__); 
	printf ("\tRestoring to Default Blocking Mode (SIGNALOFF).\n");
	regvalue = 0;
	if( ioctl( fildes, SIGNALOFF, &regvalue ) ) pioerr(__LINE__);
	printf ("SIGNALON/OFF test completed\n");

/****************************************************************************/
	/* PACKETON/OFF tests */
	printf ("\nPACKETON/OFF test\n");
	printf ("\tSetting Packet Mode (PACKETON).\n");
	regvalue = 0;
	if( ioctl( fildes, PACKETON, &regvalue ) ) pioerr(__LINE__); 
	printf ("\tRestoring to Default Non-Packet Mode (PACKETOFF).\n");
	regvalue = 0;
	if( ioctl( fildes, PACKETOFF, &regvalue ) ) pioerr(__LINE__); 
	printf ("PACKETON/OFF test completed\n");

	printf ("\nAll Ioctl(2) Call Tests Completed.\n");

} /* end of do_ioctl_test routine */
/****************************************************************************/
void
do_loadtime_test()
{
	int bytes_written;
	time_t timeval;

	printf ("\nLoad Major Time test (using ioctl calls)\n");

	/* set mode 1 (Free running mode) */
	send_packet ("A1");

	/* clear 1PPS interrupt status bit */
	regvalue = INTSTAT_1PPS;
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	printf ("\tWaiting for the 1PPS status bit...\n");

	regvalue = 0;
	do
	{
		if( ioctl( fildes, RINTSTAT, &regvalue ) ) pioerr(__LINE__); 
	}
	while (!(regvalue & INTSTAT_1PPS));

	printf ("\tGot the 1PPS status bit.\n");

	/* Clear the 1PPS status bit */
	regvalue = INTSTAT_1PPS;
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	printf ("\tLoading major time as 123 11:22:33\n");

	/* Send 123 11:22:33 as major time */
	send_packet ("B123112233");

	printf ("\tVerify that display is incrementing from 11:22:33...\n");

	sleep (10);

	printf ("\n\tNow setting to the system clock's time...\n");

	/* Get the current date and time in jjjHHMMSS format */
	timeval = time(0);

	time_buf[0] = 'B';
	num_chars = strftime (&time_buf[1], sizeof(time_buf) - 1, "%j%H%M%S", localtime(&timeval));
	if (num_chars <= 0) {
		perror ("do_loadtime_test: strftime");
		exit (1);
	}

	time_buf[num_chars + 1] = '\0';

	if (verbose)
		printf ("\tPC time buffer: %s\n", time_buf);

	printf ("\tLoading major time as %s\n", &time_buf[1]);

	/* clear 1PPS interrupt status bit */
	regvalue = INTSTAT_1PPS;
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	printf ("\tWaiting for the 1PPS status bit...\n");

	regvalue = 0;
	do
	{
		if( ioctl( fildes, RINTSTAT, &regvalue ) ) pioerr(__LINE__); 
	}
	while (!(regvalue & INTSTAT_1PPS));

	printf ("\tGot the 1PPS status bit.\n");

	/* Clear the 1PPS status bit */
	regvalue = INTSTAT_1PPS;
	if( ioctl( fildes, WINTSTAT, &regvalue ) ) pioerr(__LINE__); 

	send_packet (time_buf);

	printf ("\tVerify that display is incrementing from %.1s%.1s:%.1s%.1s:%.1s%.1s...\n",
		&time_buf[4], &time_buf[5], &time_buf[6], &time_buf[7],
		&time_buf[8], &time_buf[9]);

	sleep (10);

	printf ("Load Major Time test completed\n");
} /* end of do_loadtime_test routine */
/****************************************************************************/
/*
 *	do_GPS_test - this test sets up a loop to read the GPS/TANS packets
 *			being broadcast continuously, then injects a "Request Current
 *			Time" packet (packet ID 21) to the GPS, then filters the
 *			sugsequent packets until the response is received, namely a
 *			GPS Time packet (packet ID 41).
 *
 *			NOTE: uses globals time_requested and time_received to communicate
 *				  with rpsignalhandler
/****************************************************************************/
void
do_GPS_test()
{
	int packets_received;

	printf ("GPS Read/Write test\n");

	packets_received = 0;
	time_requested = NO;
	time_received = NO;
	setup_done = NO;

	/* set GPS mode 6 */
	send_packet ("A6");

	/* Enable signal handling. */
	regvalue = 0;
	if( ioctl( fildes, SIGNALON, &regvalue ) ) pioerr(__LINE__);

	/* Make sure we are in packet mode for the read/write calls */
	regvalue = 0;
	if( ioctl( fildes, PACKETON,&regvalue  ) ) pioerr(__LINE__); /* Enable packet mode. */

	if (verbose)
	{
		printf ("Request Current Time Packet to be sent: ");
		disp_packet (packet_21);
	}

	do
	{

		/* Receive a packet */
		read_packet();

		/* Let the read side get going before writing... */
		if (packets_received++ > 5)
		{

			/* Now inject the Request Current Time packet */
			/* unless we have already done so 			  */
			if (time_requested == NO)
			{
				time_requested = YES;
				printf ("\tRequesting Current Time from GPS...\n");
				/* Write the packet to the INFIFO */
				write_packet (packet_21, sizeof(packet_21));
			}

		} /* end of "time to send Request Current Time packet */
		
		/* Signal handler (rpsignalhandler) will look for GPS Time packet */
		/* and tell us via (global) time_received. Keep looping until we  */
		/* find it. */

	} /* end of do loop */

	while (time_received == NO);

	if (verbose)
		printf ("Disabling DPA interrupts with WMASK...\n");
	regvalue = 0;
	if( ioctl( fildes, WMASK, &regvalue ) ) pioerr(__LINE__);

	printf ("\tGPS Time packet received: ");
	disp_packet (rpacket);

	printf ("GPS Read/Write test completed\n");
	
	/* set back to mode 1 */
	send_packet("A1");
	/* Disable signal handling. */
	regvalue = 0;
	if( ioctl( fildes, SIGNALOFF, &regvalue ) ) pioerr(__LINE__);
	/* Put board back in PACKETOFF mode */
	regvalue = 0;
	if( ioctl( fildes, PACKETOFF, &regvalue ) ) pioerr(__LINE__);
	sleep(2);

} /* end of do_GPS_test routine */
/****************************************************************************/
void 
do_int_tests(test_select, mode)
int test_select;
int mode;
{
	int mask, cr0, timer1, timer2, ext_int_type;
	int all_int_tests=NO;

	if (test_select == 6)
		printf( "\n\nInterrupt tests...%s mode\n\n",
			(mode == BLOCK) ? "Blocking" : "Signaling");
	else
		printf( "\n\nInterrupt test...%s mode\n\n",
			(mode == BLOCK) ? "Blocking" : "Signaling");

next_test:

	/* Clear CR0 */
	cr0 = 0;
	prints_done = 0;
	signalcount = 0;
	printcount = 0;

	switch (test_select) 
	{
		case 0: /* No tests */
			if (verbose)
				printf ("No interrupt tests selected\n");
			return;

		case 1: /* External */
external:
			printf ("External Interrupt Test\n");
			if (interact_mode == YES)
			{
				printf ("Real (0), HB (1), or Strobe(2) external interrupt?\n");
				fscanf (stdin, "%d", &ext_int_type);
				switch (ext_int_type)
				{
					case 0:	/* No test fixture */
						printf ("Real External Interrupt Mode\n");
						break;
					case 1:	/* Heartbeat */
						/* NOTE: interrupt will be via loopback fixture */
						printf ("Heartbeat Loopback Test Interrupt Mode\n");
						timer1 = 1000;
						timer2 = 1000;
						per_second = 10000000 / (timer1 * timer2);
						printf ("\tHeartbeat will be %ld times per second\n", per_second);
						if (verbose)
							printf ("\tSetting heartbeat to %d x %d\n", timer1, timer2);
						set_heartbeat(timer1,timer2);
						cr0 = HBEN;
						break;
					case 2:	/* Strobe */
						printf ("Strobe Loopback Test Interrupt Mode\n");
						/* set strobe as about 3 seconds beyond current time...*/
						set_strobe (); 
						/* major/minor mode, which will happen exactly once */
						cr0 = STREN | STRMODE_MAJ;
						break;
				}
			}
			get_time();
			mask =  MASK_EXT;
			cr0 |=  LOCKEN | EVENTEN;
			printf ("\tWaiting for external interrupt...\n");
			break;

		case 6: /* all tests */
			all_int_tests = YES;
			if (ALL_tests == YES)
			{ /* "-A" option */
				/* Include external interrupt test */
				test_select = 1;
				goto external;
			}
			else { /* "-a" option */
				/* skip external interrupt test, */
				/* start with heartbeat test */
				test_select = 2;
				/* fall into test 2 */
			}

		case 2: /* Heartbeat */
			printf ("Heartbeat Interrupt Test\n");
			get_time();
			if (interact_mode == YES)
			{
				interact_mode = NO; /* Turn off in case of looping */
				printf ("Please enter timer 1 value: ");
				fscanf (stdin, "%d", &timer1);
				printf ("Please enter timer 2 value: ");
				fscanf (stdin, "%d", &timer2);
			}
			else
			{
				/* Default timers to be ten per second */
				timer1 = 1000;
				timer2 = 1000;
			}
			per_second = 10000000 / (timer1 * timer2);
			printf ("\tHeartbeat will be %ld times per second\n", per_second);
			if (per_second > 4500)
			{
				printf ("Heartbeat too fast...");
				printf ("Signal handler cannot keep up!\n");
				printf ("Aborting test.\n");
				return;
			}
			if (verbose)
				printf ("\tSetting heartbeat to %d x %d\n", timer1, timer2);
			set_heartbeat(timer1,timer2);
			mask =  MASK_HEART;
			cr0 = LOCKEN | HBEN;
			break;

		case 3: /* Major/minor Strobe */
			printf ("Major/minor Strobe Interrupt Test\n");
			/* set strobe as about 3 seconds beyond current time...*/
			set_strobe (); 
			mask =  MASK_STROBE;
			/* major/minor mode, which will happen exactly once */
			cr0 = LOCKEN | STREN | STRMODE_MAJ;
			break;

		case 4: /* Minor-only Strobe */
			printf ("Minor-only Strobe Interrupt Test\n");
			/* set strobe as about 3 seconds beyond current time...*/
			set_strobe (); 
			mask =  MASK_STROBE;
			/* minor-only, which will happen once a second */
			cr0 = LOCKEN | STREN | STRMODE_MIN; 
			break;

		case 5: /* 1PPS */
			printf ("1PPS Interrupt Test\n");
			mask =  MASK_1PPS;
			cr0 = LOCKEN; 
			break;

		default: /* none of the above */
			fprintf (stderr, "Invalid test_select: %d\n", test_select);
			return;

	} /* end of test_select switch */

	if ( verbose && (fast_terminal == YES) )
	{
		printf( "\n    About to make the WMASK interrupt enable call... \n" );
	}

	if (mode == SIGNAL)
	{

		switch (test_select) 
		{

		case 0: /* No tests */
			if (verbose)
				printf ("No interrupt tests selected\n");
			return;

		case 1: /* External */
			if (ext_int_type == 1)
			{
				if (verbose)
					printf ("Acting like Heartbeat test\n");
				goto heartbeat1;
			}
			/* Otherwise, act like other tests */
		case 3: /* Major/minor Strobe */
		case 4: /* Minor-only Strobe */
		case 5: /* 1PPS */
			prints_done  = 0;		/* Clear the print counter.			*/

			signal( SIGUSR1, signalhandler ); /* Link the signal to the handler.*/
			break;

		case 2: /* Heartbeat */
heartbeat1:
			signalcount = 0;		/* Clear the signal counter.		*/
			printcount  = 0;		/* Ditto the print counter.			*/
			prints_done  = 0;		/* Ditto the print counter.			*/

			signal( SIGUSR1, hbsighandler ); /* Link signal handler to signal.	*/
			break;

		case 6: /* "all of the above" not valid here */
		default:
			fprintf (stderr, "Invalid test_select: %d\n", test_select);
			return;

		} /* end of test_select switch */
	
		/* Enable signal handling. */
		regvalue = 0;
		if( ioctl( fildes, SIGNALON, &regvalue ) ) pioerr(__LINE__);

	} /* end of "if (mode == SIGNAL)" */

	/* Wait for the interrupt.*/ 

	/* Enable the interrupt */
	if( ioctl( fildes, WCR0, &cr0 ) ) pioerr(__LINE__); 

block_wait:

	/* Clear and enable the interrupt status */
	/* The WMASK call will clear the interrupt status */
	/* both before and after it occurs */

	if( ioctl( fildes, WMASK, &mask ) ) pioerr(__LINE__); 

	if (mode == SIGNAL)
	{
signal_pause:
		if (verbose)
			printf( "\n    Non-blocked waiting...\n" );

		pause( SIGUSR1 );			/* Wait for signal to come in (BSD UNIX).*/

		if( ioctl( fildes, RUNLOCK, &regvalue ) ) pioerr(__LINE__);

		/* loop 5 times for repeating tests */
		switch (test_select)
		{
			case 1: /* External Interrupt */
				if (ext_int_type == 1)
				{
					if (verbose)
						printf ("Acting like Heartbeat test\n");
					goto heartbeat2;
				}
				else
				{
					/* Otherwise, just get the event registers */
					get_event();
					break;
				}
			case 3: /* Major/minor strobe */
				get_time();
				break;
			case 2: /* Heartbeat */
heartbeat2:
				if (prints_done < 5)
				{
			signal( SIGUSR1, hbsighandler ); /* Link the signal to the handler.*/
					goto signal_pause;
				}
				break;
			case 4: /* Minor-only Strobe */
			case 5: /* 1PPS */
				get_time();
				if (++prints_done < 5)
				{
			signal( SIGUSR1, signalhandler ); /* Link the signal to the handler.*/
					goto signal_pause;
				}
				break;
			default:
				/* No looping in other tests */
				break;
		} /* end of switch on test_select */

		/* Disable signal handling. */
		regvalue = 0;
		if( ioctl( fildes, SIGNALOFF, &regvalue ) ) pioerr(__LINE__);

	} /* end of "if (mode == SIGNAL)" */
	else
	{ /* blocking mode */


		if ( (test_select == 2) /* Internal Heartbeat */
			|| ((test_select == 1) && (ext_int_type == 1)) ) /* Loopback Heartbeat */
		{
			printcount += 1;
			if (verbose)
				printf ("%d\n", printcount);

			if( printcount == per_second )
			{

				get_time();

				printcount = 0;				/* Clear the print counter.		*/
				prints_done += 1;			/* Add one to prints done 		*/

				if (verbose)
					printf ("\tPrints_done = %d\n", prints_done);
			}
		} /* end of heartbeat section */
		else
		{
			/* Read the UNLOCK reg. */
			if( ioctl( fildes, RUNLOCK, &regvalue ) ) pioerr(__LINE__);
		}

		/* loop 5 times for repeating tests */
		switch (test_select)
		{
			case 1: /* External Interrupt */
				if (ext_int_type == 1)
				{
					if (verbose)
						printf ("Acting like Heartbeat test\n");
					goto heartbeat3;
				}
				else
				{
					/* Otherwise, just get the event registers */
					get_event();
					break;
				}
			case 3: /* Major/minor strobe */
				get_time();
				break;
			case 2: /* Heartbeat */
heartbeat3:
				if (prints_done < 5)
				{
					goto block_wait;
				}
				if (verbose)
					printf( "\n    Time after all prints done is: \n" );
				get_time();
				break;
			case 4: /* Minor-only Strobe */
			case 5: /* 1PPS */
				get_time();
				if (++prints_done < 5)
				{
					goto block_wait;
				}
				break;
			default:
				/* No looping in other tests */
				break;
		} /* end of switch on test_select */


	} /* end of blocking mode section */

	switch (test_select) 
	{
		case 0: /* No tests */
			printf ("No interrupt tests selected\n");

		case 1: /* External */
			printf ("End of External Interrupt Test\n");
			break;

		case 2: /* Heartbeat */
			printf ("End of Heartbeat Interrupt Test\n");
			break;

		case 3: /* Major/minor Strobe */
			printf ("End of Major/minor Strobe Interrupt Test\n");
			break;

		case 4: /* Minor-only Strobe */
			printf ("End of Minor-only Strobe Interrupt Test\n");
			break;

		case 5: /* 1PPS */
			printf ("End of 1PPS Interrupt Test\n");
			break;

		case 6:
		default: /* none of the above */
			fprintf (stderr, "Invalid test_select: %d\n", test_select);
			return;

	} /* end of test_select switch */

	if (verbose)
		printf ("Disabling all interrupts with WMASK...\n");
	regvalue =0;
	if( ioctl( fildes, WMASK, &regvalue ) ) pioerr(__LINE__);

	regvalue = 0;
	/* Clear the CR0 register */
	if( ioctl( fildes, WCR0, &regvalue ) ) pioerr(__LINE__); 

	/* Keep looping if multiple tests were selected */
	if (all_int_tests == YES)
	{
		test_select++;
		/* test #5 is the last one... */
		if (test_select < 6)
		{
			goto next_test;
		}
	}

	if (all_int_tests == YES)
		printf( "\nEnd of Interrupt tests...%s mode\n",
			(mode == BLOCK) ? "Blocking" : "Signaling");
	else
		printf( "\nEnd of Interrupt test...%s mode\n",
			(mode == BLOCK) ? "Blocking" : "Signaling");

	return;
} /* end of do_int_tests routine */

/****************************************************************************/
void
wtsignalhandler()
{
	int bytes_written;

	if (verbose)
		printf ("Got 1PPS signal\n");

	/* Write out packet */
	/* Note: this should not block */
	bytes_written = write (fildes, time_buf, num_chars+1);
	if (bytes_written <= 0)
	{
		perror ("write in wtsighandler");
		exit (2);
	}

	if (verbose)
		printf ("Wrote Date Packet\n");

	if (verbose)
		printf ("Disabling DPA interrupts with WMASK...\n");
	regvalue = 0;
	if( ioctl( fildes, WMASK, &regvalue ) ) pioerr(__LINE__);

} /* end of wtsignalhandler routine */
/****************************************************************************/
void
write_time (mode)
int mode;
{
	int bytes_written;
	time_t timeval;

	/* Get the current date and time in jjjHHMMSS format */
	timeval = time(0);

	num_chars = strftime (time_buf, sizeof(time_buf), "%j%H%M%S", localtime(&timeval));
	if (num_chars <= 0) {
		perror ("write_time: strftime");
		exit (1);
	}

	if (verbose)
		printf ("\tPC time buffer: %s\n", time_buf);

	printf ("\tLoading major time as %s\n", time_buf);

	/* Write time to btfp in proper mode */
	if (mode == SIGNALOFF)
	{
		/* Blocking mode */
		regvalue = 0;
		if( ioctl( fildes, SIGNALOFF, &regvalue ) ) pioerr(__LINE__); 
		bytes_written = write( fildes, time_buf, num_chars+1);
		if (bytes_written <= 0)
		{
			perror ("write in write_time");
			exit (2);
		}
	}
	else 
	{
		regvalue = 0;
		/* Non-blocking mode */
		if( ioctl( fildes, SIGNALON, &regvalue ) ) pioerr(__LINE__); 
		signal( SIGUSR1, wtsignalhandler ); /* Link the signal to the handler.*/

		/* Clear, then enable 1PPS interrupt */
		regvalue = MASK_1PPS;
		if( ioctl( fildes, WMASK, &regvalue ) ) pioerr(__LINE__); 
		if (verbose)
			printf ("Waiting for 1PPS signal...\n");
		pause( SIGUSR1 );			/* Wait for signal to come in (BSD UNIX).*/
		if (verbose)
			printf ("Resumed after getting 1PPS signal...\n");
	}
} /* end of write_time routine */

/****************************************************************************/
void
hbsighandler()						/* Handle a heartbeat signal.		*/
{
	int dummy;

	signal( SIGUSR1, hbsighandler ); /* Link signal handler to signal.	*/

	signalcount += 1;				/* Add one to the counters.		*/
	printcount += 1;

	if (verbose)
		printf ("%d\n", printcount);

	if( printcount == per_second )
	{
		if ( (display_mode == 2) || (display_mode == 3) )
		{
		printf( "       Heartbeat test signal handler has counted %d signals.\n",
							signalcount );
		}

		/* Release capture lockout. */
		if( ioctl( fildes, RUNLOCK, &dummy  ) ) pioerr(__LINE__); 

		if ( (display_mode == 1) || (display_mode == 3) )
		{
			get_time();
		}

		printcount = 0;				/* Clear the print counter.			*/
		prints_done += 1;				/* Add one to prints done 			*/

	} /* end of "if( printcount == per_second )" */

} /* end of hbsighandler */
		
/****************************************************************************/
void
signalhandler()						/* Activated when SIGUSR1 is received.	*/
{
	if (verbose)
		printf( "    signalhandler(): Signal received!\n" );
	
}

/****************************************************************************/
get_time()
{
	int dummy, i;

	if( ioctl( fildes, READTIME, &ts ) ) pioerr(__LINE__);	/* Read current time.	*/

	if (fast_terminal == YES)
	{
		printf ("      READTIME words: ");
		for (i=0; i<=4; i++ )
		{
			printf( " 0x%04x ", ts.btfp_time[ i ] );	
		}
		printf ("\n");
	}

	return;
} /* end of get_time routine */
/****************************************************************************/
get_event()
{
	int dummy, i;

	if( ioctl( fildes, READEVENT, &event ) ) pioerr(__LINE__);

	if (fast_terminal == YES)
	{
		printf ("      READEVENT words: ");
		for (i=0; i<=4; i++ )
		{
			printf( " 0x%04x ", event.btfp_event[ i ] );	
		}
		printf ("\n");
	}

	/* Release capture lockout. */
	if( ioctl( fildes, RUNLOCK, &dummy  ) ) pioerr(__LINE__); 

	return;
} /* end of get_event routine */
/****************************************************************************/
pioerr(line_no)									/* Log ioctl(2) error to terminal.*/
int line_no;
{
     char *s;
     int sys_nerr;
     int errno;

	 printf( "      ERROR RETURNED ON IOCTL CALL (%d)!\n            " ,
			line_no);
         perror( s );
}									
/****************************************************************************/
printregs()									/* Print out all of the registers.*/
{
	printf( "         ID =\t\t0x%04x.\n\n", ds.btfp_id );
	printf( "         DEVICE=\t0x%04x.\n\n", ds.btfp_device );
	printf( "         STATUS=\t0x%04x.\n\n", ds.btfp_status );

	printf( "         TIMEREQ=\t0x%04x.\n\n", ds.btfp_timereq );
	printf( "         TIME0=\t\t0x%04x.\n", ds.btfp_time[ 0 ] );	
	printf( "         TIME1=\t\t0x%04x.\n", ds.btfp_time[ 1 ] );	
	printf( "         TIME2=\t\t0x%04x.\n", ds.btfp_time[ 2 ] );	
	printf( "         TIME3=\t\t0x%04x.\n\n", ds.btfp_time[ 3 ] );	
	printf( "         TIME4=\t\t0x%04x.\n\n", ds.btfp_time[ 4 ] );	

	printf( "         EVENT0=\t0x%04x.\n", ds.btfp_event[ 0 ] );	
	printf( "         EVENT1=\t0x%04x.\n", ds.btfp_event[ 1 ] );	
	printf( "         EVENT2=\t0x%04x.\n", ds.btfp_event[ 2 ] );	
	printf( "         EVENT3=\t0x%04x.\n\n", ds.btfp_event[ 3 ] );	
	printf( "         EVENT4=\t0x%04x.\n\n", ds.btfp_event[ 4 ] );	

	printf( "         UNLOCK=\t0x%04x.\n\n", ds.btfp_unlock );		
	printf( "         ACK=\t\t0x%04x.\n\n", ds.btfp_ack );		
	
	printf( "         CR0=\t\t0x%04x.\n", ds.btfp_cr0 );		

	printf( "\tSkipping IN/OUT FIFOs.\n");		

	printf( "         MASK=\t\t0x%04x.\n", ds.btfp_mask );		
	printf( "         INTSTAT=\t0x%04x.\n", ds.btfp_intstat);		
	printf( "         VECTOR=\t0x%04x.\n", ds.btfp_vector );		
	printf( "         LEVEL=\t\t0x%04x.\n", ds.btfp_level );		
	
} /* end of printregs routine */
/****************************************************************************/
void
write_packet (byte_ptr, num_bytes)
unsigned char *byte_ptr;
int num_bytes;
{
	int bytes_written;

	/* Note: this routine uses the PACKETON write(2) method of */
	/* sending a data packet */
	/* The send_packet() routine uses the byte-by-byte method */

	regvalue = 0;
	if( ioctl( fildes, PACKETON, &regvalue ) ) pioerr(__LINE__); /* Enable packet mode. */
	bytes_written = write (fildes, byte_ptr, num_bytes);
	if (bytes_written < num_bytes)
	{
		fprintf (stderr, "Error from write in write_packet: %d\n",
				bytes_written);
		exit (2);
	}
	if (verbose)
		printf ("Returned from packeton write with %d bytes!\n",
				bytes_written);

} /* end of write_packet routine */
/****************************************************************************/
void
send_packet (byte_ptr)
unsigned char *byte_ptr;
{
	register int loop_count;

	/* Note: this routine uses the byte-by-byte method of sending a packet */
	/* The write_packet() routine uses the PACKETON write(2) method */

	/* clear the INFIFO ACK bit */
	regvalue = ACK_INFIFO;
	if( ioctl( fildes, WACK, &regvalue ) ) pioerr(__LINE__);
	
	/* write the SOH */
	regvalue = SOH;
	if( ioctl( fildes, WINFIFO, &regvalue ) ) pioerr(__LINE__);	

	/* write the null-terminated packet */
	while (*byte_ptr)
	{
		regvalue = *byte_ptr;
		if( ioctl( fildes, WINFIFO, &regvalue ) ) pioerr(__LINE__);	
		byte_ptr++;
	}

	/* write the ETB */
	regvalue = ETB;
	if( ioctl( fildes, WINFIFO, &regvalue ) ) pioerr(__LINE__);	

	/* command the TFP to take action on this packet */
	regvalue = ACK_INACT;
	if( ioctl( fildes, WACK, &regvalue ) ) pioerr(__LINE__);

	/* wait for FIFO to ACK back */

	loop_count = 0;
	regvalue = 0;
	while (loop_count <= MAX_SPIN)
	{
		if( ioctl( fildes, RACK, &regvalue ) ) pioerr(__LINE__);
		if ( regvalue & ACK_INFIFO ) /* Break if action taken */
		{
			break;
		}
		else
		{
			loop_count++;
		}
	}

	if (verbose)
		printf ("Looped %d times\n", loop_count);

	regvalue = ACK_INFIFO;
	/* clear the INFIFO ACK bit */
	if( ioctl( fildes, WACK, &regvalue ) ) pioerr(__LINE__);

	if (loop_count >= MAX_SPIN)
	{
		printf ("TIMEOUT (%d) waiting for INFIFO action\n", MAX_SPIN);
	}

} /* end of send_packet routine */
/****************************************************************************/
void
read_packet()
{
	void rpsignalhandler();

	/* Enable DPA interrupts, then leave them alone */

	if (setup_done == NO)
	{
	setup_done = YES;

	signal( SIGUSR1, rpsignalhandler ); /* Link the signal to the handler.*/
	/* Clear and enable the Data Packet Available interrupt */
	/* Note: this should not block */
	if (verbose)
		printf ("Entering non-blocking WMASK...\n");
	regvalue = MASK_DPA;
	if( ioctl( fildes, WMASK, &regvalue ) ) pioerr(__LINE__);
	if (verbose)
		printf ("Returned from non-blocking WMASK...\n");

	} /* end of initial setup section */
	/* Now wait for the packet */
	pause( SIGUSR1 );				/* Wait for signal to come in (BSD UNIX).*/
} /* end of read_packet routine */
/****************************************************************************/
void
rpsignalhandler()
{
	/* bytes_read is a global, so disp_packet can decode properly */

	if (verbose)
		printf ("Got DPA signal\n");

	/* Read packet in */
	/* Note: this should not block */
	bytes_read = read (fildes, rpacket, sizeof(rpacket));
	if (bytes_read <= 0)
	{
		fprintf (stderr, "Error from read in rpsignalhandler: %d\n",
				bytes_read);
		exit (2);
	}

	if (time_requested)
	{
		/* Scan for GPS Time packet */
		/* Coming from GPS, format is: DLE ID .... DLE ETX */
		if (rpacket[1] == (unsigned char) 0x41)
		{
			if (verbose)
				printf ("GPS Time packet received!\n");
			time_received = YES;
		}
		else
		{
			if (verbose)
				printf ("%.2x \n", rpacket[1]);
		}
	}
	signal( SIGUSR1, rpsignalhandler ); /* Link the signal to the handler.*/
	if (verbose)
		printf ("Read Data Packet!\n");
} /* end of rpsignalhandler routine */
/****************************************************************************/
/*
 *	NOTE: build_packet is not currently used. It is provided as a template for
 *  an application which may want to read in packets to be sent to the GPS
 *  Sateliite Receiver from stdin. The do_GPS_test routine would have to
 *  be modified to use this read-in packet, and the rpsignalhandler routine
 *  would have to know which packet to look for.
 */
/****************************************************************************/
void
build_packet(byte_ptr)
unsigned char *byte_ptr;
{
	int index, length, c;
	unsigned char pack_buff[2*MAXPACKET], *item_ptr, byte_value;
	long hex_value;

	/* Read in character stream from stdin */
	/* Expect space or newline-separated characters of three types: */
	/*	  single characters: ASCII characters */
	/*    double characters: hex values */
	/*    triple characters: mnemonic values, such as SOH */

	index = 0;
	while ( (c = getc (stdin)) != EOF ) 
	{
		pack_buff[index++] = (unsigned char) c;
	}

	index = 0;

	/* Expect space-separated stuff, maybe on multiple lines */
	item_ptr = (unsigned char *) strtok (pack_buff, delimiters);
	do 
	{

		length = strlen (item_ptr);
		if (verbose)
			printf ("%s: Length is %d\n", item_ptr, length);

		switch (length) 
		{
			case 1: /* Single Letters */
				/* just treat it as an ASCII character */
				if ( isascii (*item_ptr) ) 
				{
					byte_value = *item_ptr;
				}
				else 
				{
					printf ("Not an ASCII character: 0x%x\n", *item_ptr);
					exit (2);
				}
				break;

			case 2: /* Hex Values */
				if ( (isxdigit (*item_ptr)) && (isxdigit (*(item_ptr+1))) )
				{
					hex_value = strtol (item_ptr, (char **) NULL, 16);
					byte_value = (unsigned char) hex_value;
				}
				else 
				{
					printf ("Not hex digits: %s\n", item_ptr);
					exit (2);
				}
				break;
			case 3: /* SOH, ETB, etc. */
				if (strncmp (item_ptr, "SOH", 3) == 0) 
				{
					byte_value = 0x01;
				}
				else if (strncmp (item_ptr, "ETB", 3) == 0) 
				{
					byte_value = 0x17;
				}
				else if (strncmp (item_ptr, "DLE", 3) == 0) 
				{
					byte_value = 0x10;
				}
				else if (strncmp (item_ptr, "ETX", 3) == 0) 
				{
					byte_value = 0x03;
				}
				else 
				{
					printf ("Unknown keyword: %s\n", item_ptr);
					exit (1);
				}
				break;

			default:
				printf ("Illegal length (%d) of item: %s\n",
						length, item_ptr);
				exit (1);
		} /* end of "switch (length)" */

		if (verbose)
			printf ("byte_value is 0x%.2x\n", byte_value);

		*byte_ptr++ = byte_value;

		item_ptr = (unsigned char *) strtok (NULL, delimiters);

	} /* end of do loop for finding new tokens */
	while (item_ptr != (unsigned char *) NULL);

	/* Ensure that packet is NULL-terminated */
	*byte_ptr = (unsigned char) NULL;

} /* end of build_packet routine */
/****************************************************************************/
void
disp_packet(byte_ptr)
unsigned char *byte_ptr;
{
	int index, column;
	unsigned char byte_value;

	column = 0;
	byte_value = *byte_ptr++;
	for (index = 0; index < bytes_read; index++)
	{
		printf ("%.2x ", byte_value);

		if (++column == 16) 
		{
			printf ("\n");
			column = 0;
		}
		byte_value = *byte_ptr++;
	} /* end of for loop to print out packet buffer */

	printf ("\n");
} /* end of disp_packet routine */
/****************************************************************************/
void
set_heartbeat (n1, n2)
int n1, n2;
{
	unsigned char packet[12];

	packet[0] = 'F';
	packet[1] = '5'; /* synchronized mode */
	sprintf (&packet[2], "%04x", n1-1);
	sprintf (&packet[6], "%04x", n2-1);
	packet[10] = (unsigned char)NULL;
	send_packet (packet);
} /* end of set_heartbeat routine */

/****************************************************************************/
void
set_strobe ()
{
	regvalue;

tryagain:;

	/* Reads time into global ts structure */
	get_time ();

	t = ts.btfp_time[ 2 ] & 0x00ff;			 	/* Isolate the seconds field.	*/
	ts.btfp_time[ 2 ] &= 0xff00;			 	/* Clear the seconds field.	*/

	if( t >= 0x07 && t <= 0x09 )
		ts.btfp_time[ 2 ] |= 0x0012;
	else if( t >= 0x17 && t <= 0x19 )
		ts.btfp_time[ 2 ] |= 0x0022;
	else if( t >= 0x27 && t <= 0x29 )
		ts.btfp_time[ 2 ] |= 0x0032;
	else if( t >= 0x37 && t <= 0x39 )
		ts.btfp_time[ 2 ] |= 0x0042;
	else if( t >= 0x47 && t <= 0x49 )
		ts.btfp_time[ 2 ] |= 0x0052;
	else if( t >= 0x56 )
	{
		sleep( 4 );
		goto tryagain;
	}
	else
	{
		t += 0x0003;				/* Just add 3 seconds to what we had.*/	
		ts.btfp_time[ 2 ] |= t; 
									/* That is what we'll wait for.		*/
	}
	
	if (fast_terminal == YES)
	{
		printf( "\n      RTIME2- Time word 2 to the strobe is 0x%04x.\n",
							ts.btfp_time[ 2 ] );								
		printf( "\n      Now writing the later time value to the strobe ... \n" );
	}

	/* Note: the driver will disable strobe output before writing */
	/* to the Strobe registers */
	if( ioctl( fildes, WRITESTROBE, &ts.btfp_time[1] ) ) 
									pioerr(__LINE__); /* Write to strobe.	*/

} /* end of set_strobe routine */
/****************************************************************************/
void
usage(name)
char	*name;
{
	printf ("usage: %s [-aAvSLRTEIDrilpbsd]\n", name);

	printf ("\ta : Run all tests, excluding tests g, b1, and s1\n");
	printf ("\tA : Run ALL tests\n");-
	printf ("\t\tNOTE: external interrupt source and GPS Satellite Receiver\n");
	printf ("\t\t      must be connected and operational\n");
	printf ("\tv : verbose mode (prints diagnostic displays to stdout)\n");
	printf ("\tS : Slow terminal (defaults to fast terminal)\n");
	printf ("\tL : Loop through all selected tests (default: no loop)\n");
	printf ("\tR : READALLREGS test. Reads all bc63{57} registers.\n");
	printf ("\tT : READTIME test.\n");
	printf ("\tE : READEVENT test.\n");
	printf ("\tI : make heartbeat/external-interrupt tests interactive\n");
	printf ("\tD : enable driver debugging\n");
	printf ("\tr : read(2)/write(2) test.\n");
	printf ("\ti : enable most ioctl(2) tests (except for interrupts)\n");
	printf ("\tl : load major time test. Uses current time from system clock.\n");
	printf ("\tg : select GPS test. Writes to INFIFO, reads from OUTFIFO.\n");
	printf ("\t\tNOTE: GPS Satellite Receiver must be connected and operational.\n");
	printf ("\tb : select block-until-interrupt test\n");
	printf ("\t\t0 : no blocking tests (DEFAULT)\n");
	printf ("\t\t1 : external interrupt test\n");
	printf ("\t\t2 : heartbeat test\n");
	printf ("\t\t3 : strobe test (major/minor)\n");
	printf ("\t\t4 : strobe test (minor-only)\n");
	printf ("\t\t5 : 1PPS test\n");
	printf ("\t\t6 : all of the above blocking tests\n");
	printf ("\ts : select signal handling test\n");
	printf ("\t\t0 : no signal handling tests (DEFAULT)\n");
	printf ("\t\t1 : external interrupt test\n");
	printf ("\t\t2 : heartbeat test\n");
	printf ("\t\t3 : strobe test (major/minor)\n");
	printf ("\t\t4 : strobe test (minor-only)\n");
	printf ("\t\t5 : 1PPS test\n");
	printf ("\t\t6 : all of the above signal handling tests\n");
	printf ("\td : select display mode when heartbeat signal received\n");
	printf ("\t\t0 : no display\n");
	printf ("\t\t1 : display time of heartbeats (DEFAULT)\n");
	printf ("\t\t2 : display count of heartbeats\n");
	printf ("\t\t3 : display time and count of heartbeats\n");

	return;
} /* end of usage routine */
