/****************************************************************************/
/*																			
	The include file for applications calling the 

		bc635VME/bc650VXI HP-UX v9.01 device driver.

	The device name is 'btfp' for 'BANCOMM time and frequency processor'.	

	Copyright (C) 1991, BANCOMM, San Jose, CA.  All rights reserved.	
                                                                      
  	Author:	Greg Dowd	
*/
/****************************************************************************/

struct btfp_device					/* bc635 module register definitions.	*/
									/* Note that unsigned short must be used!*/
{
	unsigned short 	btfp_id;		/* ID register.					*/

	unsigned short	btfp_device;	/* Device type. 				*/

	unsigned short	btfp_status;	/* Status register.				*/
#define btfp_control	btfp_status /* Control register.			*/
	
	unsigned short	btfp_gap1[ 2 ];	/* Unused address space.		*/

	unsigned short	btfp_timereq;	/* Time Request. 				*/

	unsigned short	btfp_time [5];	/* Requested Time registers.	*/
									/* XX, XX, Stat, DH */
									/* DT, DU, HT, HU */
									/* MT, MU, ST, SU */
									/* MSH, MST, MSU, USH */
									/* UST, USU, NSH, XX */
	unsigned short	btfp_event [5];	/* Requested Event registers.	*/
									/* XX, XX, Stat, DH */
									/* MT, MU, ST, SU */
									/* MSH, MST, MSU, USH */
									/* UST, USU, NSH, XX */

#define btfp_strobe1	btfp_event[1] /* Strobe part 1 */
									/* XX, XX, HT, HU */
#define btfp_strobe2	btfp_event[2] /* Strobe part 2 */
									/* MT, MU, ST, SU */
#define btfp_strobe3	btfp_event[3] /* Strobe part 3 */
									/* MSH, MST, MSU, XX */
	
	unsigned short	btfp_unlock;	/* Release Capture Lockout	*/

	unsigned short	btfp_ack;		/* Data Acknowledge 		*/

	unsigned short	btfp_cr0;		/* Control Register 0 		*/

	unsigned char	btfp_fifo[2];		/* I/O FIFO Data	 		*/
#define btfp_outfifo	btfp_fifo[1]	/* Output FIFO (from module) */
#define btfp_infifo		btfp_fifo[1]	/* Input FIFO (to module) */

	unsigned short	btfp_mask;		/* Interrupt Mask 			*/

	unsigned short	btfp_intstat;		/* Interrupt Status		*/

	unsigned short	btfp_vector;		/* Interrupt Vector		*/

	unsigned short 	btfp_level;			/* Interrupt Level		*/
	
	unsigned short	btfp_gap2[ 6 ];		/* Unused address space.*/

}; /* end of btfp_device structure */
	
/****************************************************************************/
struct btfp_time						/* Structure for reading 5 time words	*/
									/* in one ioctl(2) operation.			*/
{
	unsigned short	btfp_time[ 5 ];	/* Time words 0, 1, 2, 3, and 4. (16bit)*/
};

/****************************************************************************/
struct btfp_event						/* Structure for reading 5 event words	*/
									/* in one ioctl(2) operation.			*/
{
	unsigned short	btfp_event[ 5 ];	/* Event words 0, 1, 2, 3, and 4. (16bit)*/
};
/****************************************************************************/

struct btfp_strobe						/* Structure for writing 3 strobe words */
									/* in one ioctl(2) operation. */
{
	unsigned short	btfp_strobe[ 3 ];	/* Strobe words 1, 2, and 3. (16bit)*/
};
/****************************************************************************/
                 				/* Ioctl commands to the btfp driver.	*/

/***** Simple ioctl commands *****/

#define RIDR _IOR('X',1,unsigned int)	/* Read the ID register. */
#define RDTR _IOR('X',2,unsigned int) /* Device Type register. */

#define RSR  _IOR('X',3,unsigned int) /* Read status register. */
#define WCR  _IOW('X',4,unsigned int) /* Write Control register. */

#define RTIMEREQ  _IO('X',5) 			/* Read Time Request register. 	*/
#define RTIME0    _IOR('X',6,unsigned int)	/* Read time word zero. */
#define RTIME1    _IOR('X',7,unsigned int)	/* Read time word one. 	*/
#define RTIME2    _IOR('X',8,unsigned int)	/* Read time word two. 	*/
#define RTIME3    _IOR('X',9,unsigned int)	/* Read time word three.*/
#define RTIME4    _IOR('X',10,unsigned int)	/* Read time word four. */

#define REVENT0   _IOR('X',11,unsigned int)	/* Read event word zero.*/
#define REVENT1   _IOR('X',12,unsigned int)	/* Read event word one.	*/
#define REVENT2   _IOR('X',13,unsigned int)	/* Read event word two. */
#define REVENT3   _IOR('X',14,unsigned int)	/* Read event word three*/
#define REVENT4   _IOR('X',15,unsigned int)	/* Read event word four.*/

#define WSTROBE1  _IOW('X',16,unsigned int)	/* Read strobe word one. */
#define WSTROBE2  _IOW('X',17,unsigned int)	/* Read strobe word two. */
#define WSTROBE3  _IOW('X',18,unsigned int)	/* Read strobe word three*/

#define RUNLOCK   _IO('X',19)					/* Release Capture Lockout */

#define RACK      _IOR('X',20,unsigned int)	/* Read Acknowledge */
#define WACK      _IOW('X',21,unsigned int)	/* Write Acknowledge. */

#define RCR0      _IOR('X',22,unsigned int)	/* Read control register	 */
#define	WCR0	  _IOW('X',23,unsigned int) 	/* Write control register 	 */

#define	ROUTFIFO  _IOR('X',24,unsigned int)	/* Read from OUTFIFO */
#define	WINFIFO	  _IOW('X',25,unsigned int)	/* Write to INFIFO */

#define RMASK     _IOR('X',26,unsigned int)	/* Read Interrupt Mask. */
#define WMASK     _IOW('X',27,unsigned int)	/* Write Interrupt Mask. */

#define RINTSTAT  _IOR('X',28,unsigned int)	/* Read Interrupt Status. */
#define WINTSTAT  _IOW('X',29,unsigned int) 	/* Write Interrupt Status. */

#define RVECTOR   _IOR('X',30,unsigned int)	/* Interrupt Vector. */
									/* Only driver can write to it */

#define RLEVEL    _IOR('X',31,unsigned int)	/* Interrupt Level. */
									/* Only driver can write to it */

/***** Compound ioctl commands *****/

#define READTIME	_IOR('X',32,struct btfp_time)
									/* Read all 5 time words in one call.	*/
#define READEVENT	_IOR('X',33,struct btfp_event)
									/* Read all 5 event words in one call.	*/
#define WRITESTROBE _IOW('X',34,struct btfp_strobe)
									/* Write all 3 strobe words in one call.*/
#define	READALLREGS	_IOR('X',35,struct btfp_device)
									/* Read all btfp regs.	*/
#define	RPACKET	 36 				/* Read packet from OUTFIFO */

#define	WPACKET	 37 				/* Write packet to INFIFO */

#define SIGNALON	_IO('X',38) 	/* turn on  signal sending from driver.*/
								 	/* Also turns off blocking on interrupts.*/
#define SIGNALOFF	_IO('X',39)		/* turn off signal sending from driver.*/
									/* Also turns on blocking on interrupts.*/
#define PACKETON	_IO('X',40)		/* turn on packet-mode for driver */
#define PACKETOFF	_IO('X',41)		/* turn off packet-mode for driver */
#define DEBUGON		_IO('X',42)		/* turn on debug-mode for driver */
#define DEBUGOFF	_IO('X',43)		/* turn off debug-mode for driver */
									 

/****************************************************************************/
				/* bc635 Requested Time/Event register masks.*/

#define	TIME_REF	0x0000		/* Reference source present */
#define	EVENT_REF	0x0000		/* Reference source present */

#define	TIME_FLY	0x0010		/* Flywheeling; time code lost */
#define	EVENT_FLY	0x0010		/* Flywheeling; time code lost */

#define TIME_OSCIN	0x0000		/* Oscillator in range */
#define EVENT_OSCIN	0x0000		/* Oscillator in range */

#define TIME_OSCOUT	0x0040		/* Oscillator out of range */
#define EVENT_OSCOUT 0x0040		/* Oscillator out of range */

/****************************************************************************/
				/* bc635 ACK register masks */

#define ACK_INFIFO	0x0001		/* Input FIFO ACK Bit */
#define ACK_1PPS	0x0002		/* 1 Pulse Per Second ACK Bit */
#define ACK_DPA		0x0004		/* Output FIFO Data Packet Available Bit */
#define ACK_MORE	0x0010		/* Output FIFO "More in OUTFIFO" Flag */
					/* 0: empty, 1: more bytes */
#define ACK_CLEAR	0x0010		/* Output FIFO Clear Bit */
#define ACK_INACT	0x0080		/* Input FIFO "Take Action" Bit */
/****************************************************************************/
				/* bc635 CR0 control register masks.*/

#define	LOCKEN		0x0001		/* Enable Capture Lockout */
#define	LOCKDIS		0x0000		/* Disable Capture Lockout */

#define HBEN		0x0002		/* Enable Periodic Time Capture */
#define HBDIS		0x0000		/* Disable Periodic Time Capture */

#define EVSENSE_FALL	0x0004		/* Event Input Active Edge Select */
					/* (Falling Edge) */
#define EVSENSE_RISE	0x0000		/* Event Input Active Edge Select */
					/* (Rising Edge) */

#define EVENTEN		0x0008		/* Event Input Time Capture Enable */
#define EVENTDIS	0x0000		/* Event Input Time Capture Disable */

#define STREN		0x0010		/* Time Coincidence Strobe Output Enable */
#define STRDIS		0x0000		/* Time Coincidence Strobe Output Disable */

#define STRMODE_MIN	0x0020		/* Time Coincidence Strobe Mode Select */
					/* (Minor-only mode) */
#define STRMODE_MAJ	0x0000		/* Time Coincidence Strobe Mode Select */
					/* (Major/Minor mode) */

#define FREQSEL_10	0x0000		/* Output Clock 10 MHz */
#define FREQSEL_5	0x0040		/* Output Clock 5 MHz */
#define FREQSEL_1	0x0080		/* Output Clock 1 MHz */

/****************************************************************************/
			/* bc635 Interrupt Mask register masks.*/

#define MASK_INT0	0x0001		/* Interrupt Source 0 */
#define MASK_EXT	0x0001		/* External Event Input */

#define MASK_INT1	0x0002		/* Interrupt Source 1 */
#define MASK_HEART	0x0002		/* Periodic Pulse Output */

#define MASK_INT2	0x0004		/* Interrupt Source 2 */
#define MASK_STROBE	0x0004		/* Time Coincidence Strobe */

#define MASK_INT3	0x0008		/* Interrupt Source 3 */
#define MASK_1PPS	0x0008		/* 1 Pulse per Second (1PPS) */

#define MASK_INT4	0x0010		/* Interrupt Source 4 */
#define MASK_DPA	0x0010		/* Data Packet Available */

/****************************************************************************/
			/* bc635 Interrupt Status register masks.*/

#define INTSTAT_INT0	0x0001		/* Interrupt Source 0 */
#define INTSTAT_EXT	0x0001		/* External Event Input */

#define INTSTAT_INT1	0x0002		/* Interrupt Source 1 */
#define INTSTAT_HEART	0x0002		/* Periodic Pulse Output */

#define INTSTAT_INT2	0x0004		/* Interrupt Source 2 */
#define INTSTAT_STROBE	0x0004		/* Time Coincidence Strobe */

#define INTSTAT_INT3	0x0008		/* Interrupt Source 3 */
#define INTSTAT_1PPS	0x0008		/* 1 Pulse per Second (1PPS) */

#define INTSTAT_INT4	0x0010		/* Interrupt Source 4 */
#define INTSTAT_DPA		0x0010		/* Data Packet Available */

/****************************************************************************/
		/* bc635 Interrupt Level register masks */

#define LEVEL_DIS	0x0000		/* Interrupts disabled */
#define LEVEL_IRQ1	0x0001		/* IRQ1 */
#define LEVEL_IRQ2	0x0002		/* IRQ2 */
#define LEVEL_IRQ3	0x0003		/* IRQ3 */
#define LEVEL_IRQ4	0x0004		/* IRQ4 */
#define LEVEL_IRQ5	0x0005		/* IRQ5 */
#define LEVEL_IRQ6	0x0006		/* IRQ6 */
#define LEVEL_IRQ7	0x0007		/* IRQ7 */

/****************************************************************************/
		/* General defines */
#define	YES	1
#define NO	0
#define SOH	0x01
#define ETB	0x17
#define DLE 0x10
#define ETX 0x03
#define MAXPACKET	512			/* Maximum bytes in FIFO packet */
#define MAX_SPIN	500000			/* maximum loops to check for INFIFO action */

/****************************************************************************/
